"""Tests for :py:mod:`sphinx.ext.napoleon.docstring` module."""

from __future__ import annotations

import re
import zlib
from collections import namedtuple
from inspect import cleandoc
from itertools import product
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from sphinx.ext.intersphinx._load import load_mappings, validate_intersphinx_mapping
from sphinx.ext.napoleon import Config
from sphinx.ext.napoleon.docstring import (
    GoogleDocstring,
    NumpyDocstring,
    _convert_type_spec,
    _recombine_set_tokens,
    _token_type,
    _tokenize_type_spec,
)
from sphinx.testing.util import etree_parse

from tests.test_extensions.ext_napoleon_pep526_data_google import PEP526GoogleClass
from tests.test_extensions.ext_napoleon_pep526_data_numpy import PEP526NumpyClass

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


class NamedtupleSubclass(namedtuple('NamedtupleSubclass', ('attr1', 'attr2'))):  # NoQA: PYI024
    """Sample namedtuple subclass

    Attributes
    ----------
    attr1 : Arbitrary type
        Quick description of attr1
    attr2 : Another arbitrary type
        Quick description of attr2
    attr3 : Type

        Adds a newline after the type

    """

    # To avoid creating a dict, as a namedtuple doesn't have it:
    __slots__ = ()

    def __new__(cls, attr1, attr2=None):
        return super().__new__(cls, attr1, attr2)


class TestNamedtupleSubclass:
    def test_attributes_docstring(self):
        config = Config()
        actual = NumpyDocstring(
            cleandoc(NamedtupleSubclass.__doc__),
            config=config,
            app=None,
            what='class',
            name='NamedtupleSubclass',
            obj=NamedtupleSubclass,
        )
        expected = """\
Sample namedtuple subclass

.. attribute:: attr1

   Quick description of attr1

   :type: Arbitrary type

.. attribute:: attr2

   Quick description of attr2

   :type: Another arbitrary type

.. attribute:: attr3

   Adds a newline after the type

   :type: Type
"""

        assert str(actual) == expected


class TestInlineAttribute:
    inline_google_docstring = (
        'inline description with '
        '``a : in code``, '
        'a :ref:`reference`, '
        'a `link <https://foo.bar>`_, '
        'a :meta public:, '
        'a :meta field: value and '
        'an host:port and HH:MM strings.'
    )

    @staticmethod
    def _docstring(source):
        rst = GoogleDocstring(
            source, config=Config(), app=None, what='attribute', name='some_data', obj=0
        )
        return str(rst)

    def test_class_data_member(self):
        source = 'data member description:\n\n- a: b'
        actual = self._docstring(source).splitlines()
        assert actual == ['data member description:', '', '- a: b']

    def test_class_data_member_inline(self):
        source = f'CustomType: {self.inline_google_docstring}'
        actual = self._docstring(source).splitlines()
        assert actual == [self.inline_google_docstring, '', ':type: CustomType']

    def test_class_data_member_inline_no_type(self):
        source = self.inline_google_docstring
        actual = self._docstring(source).splitlines()
        assert actual == [source]

    def test_class_data_member_inline_ref_in_type(self):
        source = f':py:class:`int`: {self.inline_google_docstring}'
        actual = self._docstring(source).splitlines()
        assert actual == [self.inline_google_docstring, '', ':type: :py:class:`int`']


class TestGoogleDocstring:
    docstrings = [
        (
            """Single line summary""",
            """Single line summary""",
        ),
        (
            """
        Single line summary

        Extended description

        """,
            """
        Single line summary

        Extended description
        """,
        ),
        (
            """
        Single line summary

        Args:
          arg1(str):Extended
            description of arg1
        """,
            """
        Single line summary

        :Parameters: **arg1** (*str*) -- Extended
                     description of arg1
        """,
        ),
        (
            """
        Single line summary

        Args:
          arg1(str):Extended
            description of arg1
          arg2 ( int ) : Extended
            description of arg2

        Keyword Args:
          kwarg1(str):Extended
            description of kwarg1
          kwarg2 ( int ) : Extended
            description of kwarg2""",
            """
        Single line summary

        :Parameters: * **arg1** (*str*) -- Extended
                       description of arg1
                     * **arg2** (*int*) -- Extended
                       description of arg2

        :Keyword Arguments: * **kwarg1** (*str*) -- Extended
                              description of kwarg1
                            * **kwarg2** (*int*) -- Extended
                              description of kwarg2
        """,
        ),
        (
            """
        Single line summary

        Arguments:
          arg1(str):Extended
            description of arg1
          arg2 ( int ) : Extended
            description of arg2

        Keyword Arguments:
          kwarg1(str):Extended
            description of kwarg1
          kwarg2 ( int ) : Extended
            description of kwarg2""",
            """
        Single line summary

        :Parameters: * **arg1** (*str*) -- Extended
                       description of arg1
                     * **arg2** (*int*) -- Extended
                       description of arg2

        :Keyword Arguments: * **kwarg1** (*str*) -- Extended
                              description of kwarg1
                            * **kwarg2** (*int*) -- Extended
                              description of kwarg2
        """,
        ),
        (
            """
        Single line summary

        Return:
          str:Extended
          description of return value
        """,
            """
        Single line summary

        :returns: *str* -- Extended
                  description of return value
        """,
        ),
        (
            """
        Single line summary

        Returns:
          str:Extended
          description of return value
        """,
            """
        Single line summary

        :returns: *str* -- Extended
                  description of return value
        """,
        ),
        (
            """
        Single line summary

        Returns:
          Extended
          description of return value
        """,
            """
        Single line summary

        :returns: Extended
                  description of return value
        """,
        ),
        (
            """
        Single line summary

        Returns:
          Extended
        """,
            """
        Single line summary

        :returns: Extended
        """,
        ),
        (
            """
        Single line summary

        Args:
          arg1(str):Extended
            description of arg1
          *args: Variable length argument list.
          **kwargs: Arbitrary keyword arguments.
        """,
            """
        Single line summary

        :Parameters: * **arg1** (*str*) -- Extended
                       description of arg1
                     * **\\*args** -- Variable length argument list.
                     * **\\*\\*kwargs** -- Arbitrary keyword arguments.
        """,
        ),
        (
            """
        Single line summary

        Args:
          arg1 (list(int)): Description
          arg2 (list[int]): Description
          arg3 (dict(str, int)): Description
          arg4 (dict[str, int]): Description
        """,
            """
        Single line summary

        :Parameters: * **arg1** (*list(int)*) -- Description
                     * **arg2** (*list[int]*) -- Description
                     * **arg3** (*dict(str, int)*) -- Description
                     * **arg4** (*dict[str, int]*) -- Description
        """,
        ),
        (
            """
        Single line summary

        Receive:
          arg1 (list(int)): Description
          arg2 (list[int]): Description
        """,
            """
        Single line summary

        :Receives: * **arg1** (*list(int)*) -- Description
                   * **arg2** (*list[int]*) -- Description
        """,
        ),
        (
            """
        Single line summary

        Receives:
          arg1 (list(int)): Description
          arg2 (list[int]): Description
        """,
            """
        Single line summary

        :Receives: * **arg1** (*list(int)*) -- Description
                   * **arg2** (*list[int]*) -- Description
        """,
        ),
        (
            """
        Single line summary

        Yield:
          str:Extended
          description of yielded value
        """,
            """
        Single line summary

        :Yields: *str* -- Extended
                 description of yielded value
        """,
        ),
        (
            """
        Single line summary

        Yields:
          Extended
          description of yielded value
        """,
            """
        Single line summary

        :Yields: Extended
                 description of yielded value
        """,
        ),
        (
            """
        Single line summary

        Args:

          arg1 (list of str): Extended
              description of arg1.
          arg2 (tuple of int): Extended
              description of arg2.
          arg3 (tuple of list of float): Extended
              description of arg3.
          arg4 (int, float, or list of bool): Extended
              description of arg4.
          arg5 (list of int, float, or bool): Extended
              description of arg5.
          arg6 (list of int or float): Extended
              description of arg6.
        """,
            """
        Single line summary

        :Parameters: * **arg1** (*list of str*) -- Extended
                       description of arg1.
                     * **arg2** (*tuple of int*) -- Extended
                       description of arg2.
                     * **arg3** (*tuple of list of float*) -- Extended
                       description of arg3.
                     * **arg4** (*int, float, or list of bool*) -- Extended
                       description of arg4.
                     * **arg5** (*list of int, float, or bool*) -- Extended
                       description of arg5.
                     * **arg6** (*list of int or float*) -- Extended
                       description of arg6.
        """,
        ),
    ]

    def test_sphinx_admonitions(self):
        admonition_map = {
            'Attention': 'attention',
            'Caution': 'caution',
            'Danger': 'danger',
            'Error': 'error',
            'Hint': 'hint',
            'Important': 'important',
            'Note': 'note',
            'Tip': 'tip',
            'Todo': 'todo',
            'Warning': 'warning',
            'Warnings': 'warning',
        }
        config = Config()
        for section, admonition in admonition_map.items():
            # Multiline
            actual = GoogleDocstring(
                f'{section}:\n'
                '    this is the first line\n'
                '\n'
                '    and this is the second line\n',
                config,
            )
            expect = (
                f'.. {admonition}::\n'
                '\n'
                '   this is the first line\n'
                '   \n'
                '   and this is the second line\n'
            )
            assert str(actual) == expect

            # Single line
            actual = GoogleDocstring(f'{section}:\n    this is a single line\n', config)
            expect = f'.. {admonition}:: this is a single line\n'
            assert str(actual) == expect

    def test_docstrings(self):
        config = Config(
            napoleon_use_param=False,
            napoleon_use_rtype=False,
            napoleon_use_keyword=False,
        )
        for docstring, expected in self.docstrings:
            actual = GoogleDocstring(dedent(docstring), config)
            expected = dedent(expected)
            assert str(actual) == expected

    def test_parameters_with_class_reference(self):
        docstring = """\
Construct a new XBlock.

This class should only be used by runtimes.

Arguments:
    runtime (:py:class:`~typing.Dict`\\[:py:class:`int`,:py:class:`str`\\]): Use it to
        access the environment. It is available in XBlock code
        as ``self.runtime``.

    field_data (:py:class:`FieldData`): Interface used by the XBlock
        fields to access their data from wherever it is persisted.

    scope_ids (:py:class:`ScopeIds`): Identifiers needed to resolve scopes.

"""

        actual = GoogleDocstring(docstring)
        expected = """\
Construct a new XBlock.

This class should only be used by runtimes.

:param runtime: Use it to
                access the environment. It is available in XBlock code
                as ``self.runtime``.
:type runtime: :py:class:`~typing.Dict`\\[:py:class:`int`,:py:class:`str`\\]
:param field_data: Interface used by the XBlock
                   fields to access their data from wherever it is persisted.
:type field_data: :py:class:`FieldData`
:param scope_ids: Identifiers needed to resolve scopes.
:type scope_ids: :py:class:`ScopeIds`
"""
        assert str(actual) == expected

    def test_attributes_with_class_reference(self):
        docstring = """\
Attributes:
    in_attr(:py:class:`numpy.ndarray`): super-dooper attribute
"""

        actual = GoogleDocstring(docstring)
        expected = """\
.. attribute:: in_attr

   super-dooper attribute

   :type: :py:class:`numpy.ndarray`
"""
        assert str(actual) == expected

        docstring = """\
Attributes:
    in_attr(numpy.ndarray): super-dooper attribute
"""

        actual = GoogleDocstring(docstring)
        expected = """\
.. attribute:: in_attr

   super-dooper attribute

   :type: numpy.ndarray
"""

    def test_attributes_with_use_ivar(self):
        docstring = """\
Attributes:
    foo (int): blah blah
    bar (str): blah blah
"""

        config = Config(napoleon_use_ivar=True)
        actual = GoogleDocstring(docstring, config, obj=self.__class__)
        expected = """\
:ivar foo: blah blah
:vartype foo: int
:ivar bar: blah blah
:vartype bar: str
"""
        assert str(actual) == expected

    def test_code_block_in_returns_section(self):
        docstring = """
Returns:
    foobar: foo::

        codecode
        codecode
"""
        expected = """
:returns:

          foo::

              codecode
              codecode
:rtype: foobar
"""
        actual = GoogleDocstring(docstring)
        assert str(actual) == expected

    def test_colon_in_return_type(self):
        docstring = """Example property.

Returns:
    :py:class:`~.module.submodule.SomeClass`: an example instance
    if available, None if not available.
"""
        expected = """Example property.

:returns: an example instance
          if available, None if not available.
:rtype: :py:class:`~.module.submodule.SomeClass`
"""
        actual = GoogleDocstring(docstring)
        assert str(actual) == expected

    def test_xrefs_in_return_type(self):
        docstring = """Example Function

Returns:
    :py:class:`numpy.ndarray`: A :math:`n \\times 2` array containing
    a bunch of math items
"""
        expected = """Example Function

:returns: A :math:`n \\times 2` array containing
          a bunch of math items
:rtype: :py:class:`numpy.ndarray`
"""
        actual = GoogleDocstring(docstring)
        assert str(actual) == expected

    def test_raises_types(self):
        docstrings = [
            (
                """
Example Function

Raises:
    RuntimeError:
        A setting wasn't specified, or was invalid.
    ValueError:
        Something something value error.
    :py:class:`AttributeError`
        errors for missing attributes.
    ~InvalidDimensionsError
        If the dimensions couldn't be parsed.
    `InvalidArgumentsError`
        If the arguments are invalid.
    :py:exc:`~ValueError`
        If the arguments are wrong.

""",
                """
Example Function

:raises RuntimeError: A setting wasn't specified, or was invalid.
:raises ValueError: Something something value error.
:raises AttributeError: errors for missing attributes.
:raises ~InvalidDimensionsError: If the dimensions couldn't be parsed.
:raises InvalidArgumentsError: If the arguments are invalid.
:raises ~ValueError: If the arguments are wrong.
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    InvalidDimensionsError

""",
                """
Example Function

:raises InvalidDimensionsError:
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    Invalid Dimensions Error

""",
                """
Example Function

:raises Invalid Dimensions Error:
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    Invalid Dimensions Error: With description

""",
                """
Example Function

:raises Invalid Dimensions Error: With description
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    InvalidDimensionsError: If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises InvalidDimensionsError: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    Invalid Dimensions Error: If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises Invalid Dimensions Error: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises If the dimensions couldn't be parsed.:
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    :py:class:`exc.InvalidDimensionsError`

""",
                """
Example Function

:raises exc.InvalidDimensionsError:
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    :py:class:`exc.InvalidDimensionsError`: If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises exc.InvalidDimensionsError: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    :py:class:`exc.InvalidDimensionsError`: If the dimensions couldn't be parsed,
       then a :py:class:`exc.InvalidDimensionsError` will be raised.

""",
                """
Example Function

:raises exc.InvalidDimensionsError: If the dimensions couldn't be parsed,
    then a :py:class:`exc.InvalidDimensionsError` will be raised.
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    :py:class:`exc.InvalidDimensionsError`: If the dimensions couldn't be parsed.
    :py:class:`exc.InvalidArgumentsError`: If the arguments are invalid.

""",
                """
Example Function

:raises exc.InvalidDimensionsError: If the dimensions couldn't be parsed.
:raises exc.InvalidArgumentsError: If the arguments are invalid.
""",
            ),
            ################################
            (
                """
Example Function

Raises:
    :py:class:`exc.InvalidDimensionsError`
    :py:class:`exc.InvalidArgumentsError`

""",
                """
Example Function

:raises exc.InvalidDimensionsError:
:raises exc.InvalidArgumentsError:
""",
            ),
        ]
        for docstring, expected in docstrings:
            actual = GoogleDocstring(docstring)
            assert str(actual) == expected

    def test_kwargs_in_arguments(self):
        docstring = """Allows to create attributes binded to this device.

Some other paragraph.

Code sample for usage::

  dev.bind(loopback=Loopback)
  dev.loopback.configure()

Arguments:
  **kwargs: name/class pairs that will create resource-managers
    bound as instance attributes to this instance. See code
    example above.
"""
        expected = """Allows to create attributes binded to this device.

Some other paragraph.

Code sample for usage::

  dev.bind(loopback=Loopback)
  dev.loopback.configure()

:param \\*\\*kwargs: name/class pairs that will create resource-managers
                   bound as instance attributes to this instance. See code
                   example above.
"""
        actual = GoogleDocstring(docstring)
        assert str(actual) == expected

    def test_section_header_formatting(self):
        docstrings = [
            (
                """
Summary line

Example:
    Multiline reStructuredText
    literal code block

""",
                """
Summary line

.. rubric:: Example

Multiline reStructuredText
literal code block
""",
            ),
            ################################
            (
                """
Summary line

Example::

    Multiline reStructuredText
    literal code block

""",
                """
Summary line

Example::

    Multiline reStructuredText
    literal code block
""",
            ),
            ################################
            (
                """
Summary line

:Example:

    Multiline reStructuredText
    literal code block

""",
                """
Summary line

:Example:

    Multiline reStructuredText
    literal code block
""",
            ),
        ]
        for docstring, expected in docstrings:
            actual = GoogleDocstring(docstring)
            assert str(actual) == expected

    def test_list_in_parameter_description(self):
        docstring = """One line summary.

Parameters:
    no_list (int):
    one_bullet_empty (int):
        *
    one_bullet_single_line (int):
        - first line
    one_bullet_two_lines (int):
        +   first line
            continued
    two_bullets_single_line (int):
        -  first line
        -  second line
    two_bullets_two_lines (int):
        * first line
          continued
        * second line
          continued
    one_enumeration_single_line (int):
        1.  first line
    one_enumeration_two_lines (int):
        1)   first line
             continued
    two_enumerations_one_line (int):
        (iii) first line
        (iv) second line
    two_enumerations_two_lines (int):
        a. first line
           continued
        b. second line
           continued
    one_definition_one_line (int):
        item 1
            first line
    one_definition_two_lines (int):
        item 1
            first line
            continued
    two_definitions_one_line (int):
        item 1
            first line
        item 2
            second line
    two_definitions_two_lines (int):
        item 1
            first line
            continued
        item 2
            second line
            continued
    one_definition_blank_line (int):
        item 1

            first line

            extra first line

    two_definitions_blank_lines (int):
        item 1

            first line

            extra first line

        item 2

            second line

            extra second line

    definition_after_inline_text (int): text line

        item 1
            first line

    definition_after_normal_text (int):
        text line

        item 1
            first line
"""

        expected = """One line summary.

:param no_list:
:type no_list: int
:param one_bullet_empty:
                         *
:type one_bullet_empty: int
:param one_bullet_single_line:
                               - first line
:type one_bullet_single_line: int
:param one_bullet_two_lines:
                             +   first line
                                 continued
:type one_bullet_two_lines: int
:param two_bullets_single_line:
                                -  first line
                                -  second line
:type two_bullets_single_line: int
:param two_bullets_two_lines:
                              * first line
                                continued
                              * second line
                                continued
:type two_bullets_two_lines: int
:param one_enumeration_single_line:
                                    1.  first line
:type one_enumeration_single_line: int
:param one_enumeration_two_lines:
                                  1)   first line
                                       continued
:type one_enumeration_two_lines: int
:param two_enumerations_one_line:
                                  (iii) first line
                                  (iv) second line
:type two_enumerations_one_line: int
:param two_enumerations_two_lines:
                                   a. first line
                                      continued
                                   b. second line
                                      continued
:type two_enumerations_two_lines: int
:param one_definition_one_line:
                                item 1
                                    first line
:type one_definition_one_line: int
:param one_definition_two_lines:
                                 item 1
                                     first line
                                     continued
:type one_definition_two_lines: int
:param two_definitions_one_line:
                                 item 1
                                     first line
                                 item 2
                                     second line
:type two_definitions_one_line: int
:param two_definitions_two_lines:
                                  item 1
                                      first line
                                      continued
                                  item 2
                                      second line
                                      continued
:type two_definitions_two_lines: int
:param one_definition_blank_line:
                                  item 1

                                      first line

                                      extra first line
:type one_definition_blank_line: int
:param two_definitions_blank_lines:
                                    item 1

                                        first line

                                        extra first line

                                    item 2

                                        second line

                                        extra second line
:type two_definitions_blank_lines: int
:param definition_after_inline_text: text line

                                     item 1
                                         first line
:type definition_after_inline_text: int
:param definition_after_normal_text: text line

                                     item 1
                                         first line
:type definition_after_normal_text: int
"""
        config = Config(napoleon_use_param=True)
        actual = GoogleDocstring(docstring, config)
        assert str(actual) == expected

        expected = """One line summary.

:Parameters: * **no_list** (*int*)
             * **one_bullet_empty** (*int*) --

               *
             * **one_bullet_single_line** (*int*) --

               - first line
             * **one_bullet_two_lines** (*int*) --

               +   first line
                   continued
             * **two_bullets_single_line** (*int*) --

               -  first line
               -  second line
             * **two_bullets_two_lines** (*int*) --

               * first line
                 continued
               * second line
                 continued
             * **one_enumeration_single_line** (*int*) --

               1.  first line
             * **one_enumeration_two_lines** (*int*) --

               1)   first line
                    continued
             * **two_enumerations_one_line** (*int*) --

               (iii) first line
               (iv) second line
             * **two_enumerations_two_lines** (*int*) --

               a. first line
                  continued
               b. second line
                  continued
             * **one_definition_one_line** (*int*) --

               item 1
                   first line
             * **one_definition_two_lines** (*int*) --

               item 1
                   first line
                   continued
             * **two_definitions_one_line** (*int*) --

               item 1
                   first line
               item 2
                   second line
             * **two_definitions_two_lines** (*int*) --

               item 1
                   first line
                   continued
               item 2
                   second line
                   continued
             * **one_definition_blank_line** (*int*) --

               item 1

                   first line

                   extra first line
             * **two_definitions_blank_lines** (*int*) --

               item 1

                   first line

                   extra first line

               item 2

                   second line

                   extra second line
             * **definition_after_inline_text** (*int*) -- text line

               item 1
                   first line
             * **definition_after_normal_text** (*int*) -- text line

               item 1
                   first line
"""
        config = Config(napoleon_use_param=False)
        actual = GoogleDocstring(docstring, config)
        assert str(actual) == expected

    def test_custom_generic_sections(self):
        docstrings = (
            (
                """\
Really Important Details:
    You should listen to me!
""",
                """.. rubric:: Really Important Details

You should listen to me!
""",
            ),
            (
                """\
Sooper Warning:
    Stop hitting yourself!
""",
                """:Warns: **Stop hitting yourself!**
""",
            ),
            (
                """\
Params Style:
    arg1 (int): Description of arg1
    arg2 (str): Description of arg2

""",
                """\
:Params Style: * **arg1** (*int*) -- Description of arg1
               * **arg2** (*str*) -- Description of arg2
""",
            ),
            (
                """\
Returns Style:
    description of custom section

""",
                """:Returns Style: description of custom section
""",
            ),
        )

        test_config = Config(
            napoleon_custom_sections=[
                'Really Important Details',
                ('Sooper Warning', 'warns'),
                ('Params Style', 'params_style'),
                ('Returns Style', 'returns_style'),
            ]
        )

        for docstring, expected in docstrings:
            actual = GoogleDocstring(docstring, test_config)
            assert str(actual) == expected

    def test_noindex(self):
        docstring = """
Attributes:
    arg
        description

Methods:
    func(i, j)
        description
"""

        expected = """
.. attribute:: arg
   :no-index:

   description

.. method:: func(i, j)
   :no-index:

   
   description
"""  # NoQA: W293
        config = Config()
        actual = GoogleDocstring(
            docstring,
            config=config,
            app=None,
            what='module',
            options={'no-index': True},
        )
        assert str(actual) == expected

    def test_keywords_with_types(self):
        docstring = """\
Do as you please

Keyword Args:
    gotham_is_yours (None): shall interfere.
"""
        actual = GoogleDocstring(docstring)
        expected = """\
Do as you please

:keyword gotham_is_yours: shall interfere.
:kwtype gotham_is_yours: None
"""
        assert str(actual) == expected

    def test_pep526_annotations(self):
        # Test class attributes annotations
        config = Config(
            napoleon_attr_annotations=True,
        )
        actual = GoogleDocstring(
            cleandoc(PEP526GoogleClass.__doc__),
            config,
            app=None,
            what='class',
            obj=PEP526GoogleClass,
        )
        expected = """\
Sample class with PEP 526 annotations and google docstring.

.. attribute:: attr1

   Attr1 description.

   :type: int

.. attribute:: attr2

   Attr2 description.

   :type: str
"""
        assert str(actual) == expected

    def test_preprocess_types(self):
        docstring = """\
Do as you please

Yield:
   str:Extended
"""
        actual = GoogleDocstring(docstring)
        expected = """\
Do as you please

:Yields: *str* -- Extended
"""
        assert str(actual) == expected

        config = Config(napoleon_preprocess_types=True)
        actual = GoogleDocstring(docstring, config)
        expected = """\
Do as you please

:Yields: :py:class:`str` -- Extended
"""
        assert str(actual) == expected


class TestNumpyDocstring:
    docstrings = [
        (
            """Single line summary""",
            """Single line summary""",
        ),
        (
            """
        Single line summary

        Extended description

        """,
            """
        Single line summary

        Extended description
        """,
        ),
        (
            """
        Single line summary

        Parameters
        ----------
        arg1:str
            Extended
            description of arg1
        """,
            """
        Single line summary

        :Parameters: **arg1** (:py:class:`str`) -- Extended
                     description of arg1
        """,
        ),
        (
            """
        Single line summary

        Parameters
        ----------
        arg1:str
            Extended
            description of arg1
        arg2 : int
            Extended
            description of arg2

        Keyword Arguments
        -----------------
          kwarg1:str
              Extended
              description of kwarg1
          kwarg2 : int
              Extended
              description of kwarg2
        """,
            """
        Single line summary

        :Parameters: * **arg1** (:py:class:`str`) -- Extended
                       description of arg1
                     * **arg2** (:py:class:`int`) -- Extended
                       description of arg2

        :Keyword Arguments: * **kwarg1** (:py:class:`str`) -- Extended
                              description of kwarg1
                            * **kwarg2** (:py:class:`int`) -- Extended
                              description of kwarg2
        """,
        ),
        (
            """
        Single line summary

        Return
        ------
        str
            Extended
            description of return value
        """,
            """
        Single line summary

        :returns: :py:class:`str` -- Extended
                  description of return value
        """,
        ),
        (
            """
        Single line summary

        Returns
        -------
        str
            Extended
            description of return value
        """,
            """
        Single line summary

        :returns: :py:class:`str` -- Extended
                  description of return value
        """,
        ),
        (
            """
        Single line summary

        Parameters
        ----------
        arg1:str
             Extended description of arg1
        *args:
            Variable length argument list.
        **kwargs:
            Arbitrary keyword arguments.
        """,
            """
        Single line summary

        :Parameters: * **arg1** (:py:class:`str`) -- Extended description of arg1
                     * **\\*args** -- Variable length argument list.
                     * **\\*\\*kwargs** -- Arbitrary keyword arguments.
        """,
        ),
        (
            """
        Single line summary

        Parameters
        ----------
        arg1:str
             Extended description of arg1
        *args, **kwargs:
            Variable length argument list and arbitrary keyword arguments.
        """,
            """
        Single line summary

        :Parameters: * **arg1** (:py:class:`str`) -- Extended description of arg1
                     * **\\*args, \\*\\*kwargs** -- Variable length argument list and arbitrary keyword arguments.
        """,
        ),
        (
            """
        Single line summary

        Receive
        -------
        arg1:str
            Extended
            description of arg1
        arg2 : int
            Extended
            description of arg2
        """,
            """
        Single line summary

        :Receives: * **arg1** (:py:class:`str`) -- Extended
                     description of arg1
                   * **arg2** (:py:class:`int`) -- Extended
                     description of arg2
        """,
        ),
        (
            """
        Single line summary

        Receives
        --------
        arg1:str
            Extended
            description of arg1
        arg2 : int
            Extended
            description of arg2
        """,
            """
        Single line summary

        :Receives: * **arg1** (:py:class:`str`) -- Extended
                     description of arg1
                   * **arg2** (:py:class:`int`) -- Extended
                     description of arg2
        """,
        ),
        (
            """
        Single line summary

        Yield
        -----
        str
            Extended
            description of yielded value
        """,
            """
        Single line summary

        :Yields: :py:class:`str` -- Extended
                 description of yielded value
        """,
        ),
        (
            """
        Single line summary

        Yields
        ------
        str
            Extended
            description of yielded value
        """,
            """
        Single line summary

        :Yields: :py:class:`str` -- Extended
                 description of yielded value
        """,
        ),
    ]

    def test_sphinx_admonitions(self):
        admonition_map = {
            'Attention': 'attention',
            'Caution': 'caution',
            'Danger': 'danger',
            'Error': 'error',
            'Hint': 'hint',
            'Important': 'important',
            'Note': 'note',
            'Tip': 'tip',
            'Todo': 'todo',
            'Warning': 'warning',
            'Warnings': 'warning',
        }
        config = Config()
        for section, admonition in admonition_map.items():
            # Multiline
            underline = '-' * len(section)
            actual = NumpyDocstring(
                f'{section}\n'
                f'{underline}\n'
                '    this is the first line\n'
                '\n'
                '    and this is the second line\n',
                config,
            )
            expect = (
                f'.. {admonition}::\n'
                '\n'
                '   this is the first line\n'
                '   \n'
                '   and this is the second line\n'
            )
            assert str(actual) == expect

            # Single line
            actual = NumpyDocstring(
                f'{section}\n{"-" * len(section)}\n    this is a single line\n',
                config,
            )
            expect = f'.. {admonition}:: this is a single line\n'
            assert str(actual) == expect

    def test_docstrings(self):
        config = Config(
            napoleon_use_param=False,
            napoleon_use_rtype=False,
            napoleon_use_keyword=False,
            napoleon_preprocess_types=True,
        )
        for docstring, expected in self.docstrings:
            actual = NumpyDocstring(dedent(docstring), config)
            expected = dedent(expected)
            assert str(actual) == expected

    def test_type_preprocessor(self):
        docstring = dedent("""
        Single line summary

        Parameters
        ----------
        arg1:str
            Extended
            description of arg1
        """)

        config = Config(napoleon_preprocess_types=False, napoleon_use_param=False)
        actual = NumpyDocstring(docstring, config)
        expected = dedent("""
        Single line summary

        :Parameters: **arg1** (*str*) -- Extended
                     description of arg1
        """)
        assert str(actual) == expected

    def test_parameters_with_class_reference(self):
        docstring = """\
Parameters
----------
param1 : :class:`MyClass <name.space.MyClass>` instance

Other Parameters
----------------
param2 : :class:`MyClass <name.space.MyClass>` instance

"""

        config = Config(napoleon_use_param=False)
        actual = NumpyDocstring(docstring, config)
        expected = """\
:Parameters: **param1** (:class:`MyClass <name.space.MyClass>` instance)

:Other Parameters: **param2** (:class:`MyClass <name.space.MyClass>` instance)
"""
        assert str(actual) == expected

        config = Config(napoleon_use_param=True)
        actual = NumpyDocstring(docstring, config)
        expected = """\
:param param1:
:type param1: :class:`MyClass <name.space.MyClass>` instance

:param param2:
:type param2: :class:`MyClass <name.space.MyClass>` instance
"""
        assert str(actual) == expected

    def test_multiple_parameters(self):
        docstring = """\
Parameters
----------
x1, x2 : array_like
    Input arrays, description of ``x1``, ``x2``.

"""

        config = Config(napoleon_use_param=False)
        actual = NumpyDocstring(docstring, config)
        expected = """\
:Parameters: **x1, x2** (*array_like*) -- Input arrays, description of ``x1``, ``x2``.
"""
        assert str(actual) == expected

        config = Config(napoleon_use_param=True)
        actual = NumpyDocstring(dedent(docstring), config)
        expected = """\
:param x1: Input arrays, description of ``x1``, ``x2``.
:type x1: array_like
:param x2: Input arrays, description of ``x1``, ``x2``.
:type x2: array_like
"""
        assert str(actual) == expected

    def test_parameters_without_class_reference(self):
        docstring = """\
Parameters
----------
param1 : MyClass instance

"""

        config = Config(napoleon_use_param=False)
        actual = NumpyDocstring(docstring, config)
        expected = """\
:Parameters: **param1** (*MyClass instance*)
"""
        assert str(actual) == expected

        config = Config(napoleon_use_param=True)
        actual = NumpyDocstring(dedent(docstring), config)
        expected = """\
:param param1:
:type param1: MyClass instance
"""
        assert str(actual) == expected

    def test_see_also_refs(self):
        docstring = """\
numpy.multivariate_normal(mean, cov, shape=None, spam=None)

See Also
--------
some, other, funcs
otherfunc : relationship

"""

        actual = NumpyDocstring(docstring)

        expected = """\
numpy.multivariate_normal(mean, cov, shape=None, spam=None)

.. seealso::

   :py:obj:`some`, :py:obj:`other`, :py:obj:`funcs`
   \n\
   :py:obj:`otherfunc`
       relationship
"""
        assert str(actual) == expected

        docstring = """\
numpy.multivariate_normal(mean, cov, shape=None, spam=None)

See Also
--------
some, other, funcs
otherfunc : relationship

"""

        config = Config()
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'method')

        expected = """\
numpy.multivariate_normal(mean, cov, shape=None, spam=None)

.. seealso::

   :py:obj:`some`, :py:obj:`other`, :py:obj:`funcs`
   \n\
   :py:obj:`otherfunc`
       relationship
"""
        assert str(actual) == expected

        docstring = """\
numpy.multivariate_normal(mean, cov, shape=None, spam=None)

See Also
--------
some, other, :func:`funcs`
otherfunc : relationship

"""
        translations = {
            'other': 'MyClass.other',
            'otherfunc': ':func:`~my_package.otherfunc`',
        }
        config = Config(napoleon_type_aliases=translations)
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'method')

        expected = """\
numpy.multivariate_normal(mean, cov, shape=None, spam=None)

.. seealso::

   :py:obj:`some`, :py:obj:`MyClass.other`, :func:`funcs`
   \n\
   :func:`~my_package.otherfunc`
       relationship
"""
        assert str(actual) == expected

    def test_colon_in_return_type(self):
        docstring = """
Summary

Returns
-------
:py:class:`~my_mod.my_class`
    an instance of :py:class:`~my_mod.my_class`
"""

        expected = """
Summary

:returns: an instance of :py:class:`~my_mod.my_class`
:rtype: :py:class:`~my_mod.my_class`
"""

        config = Config()
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'method')

        assert str(actual) == expected

    def test_underscore_in_attribute(self):
        docstring = """
Attributes
----------

arg_ : type
    some description
"""

        expected = """
:ivar arg_: some description
:vartype arg_: type
"""

        config = Config(napoleon_use_ivar=True)
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'class')

        assert str(actual) == expected

    def test_underscore_in_attribute_strip_signature_backslash(self):
        docstring = """
Attributes
----------

arg_ : type
    some description
"""

        expected = """
:ivar arg\\_: some description
:vartype arg\\_: type
"""

        config = Config(napoleon_use_ivar=True)
        config.strip_signature_backslash = True
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'class')

        assert str(actual) == expected

    def test_return_types(self):
        docstring = dedent("""
            Returns
            -------
            DataFrame
                a dataframe
        """)
        expected = dedent("""
           :returns: a dataframe
           :rtype: :py:class:`~pandas.DataFrame`
        """)
        translations = {
            'DataFrame': '~pandas.DataFrame',
        }
        config = Config(
            napoleon_use_param=True,
            napoleon_use_rtype=True,
            napoleon_preprocess_types=True,
            napoleon_type_aliases=translations,
        )
        actual = NumpyDocstring(docstring, config)
        assert str(actual) == expected

    def test_yield_types(self):
        docstring = dedent("""
            Example Function

            Yields
            ------
            scalar or array-like
                The result of the computation
        """)
        expected = dedent("""
            Example Function

            :Yields: :term:`scalar` or :py:class:`array-like <numpy.ndarray>` -- The result of the computation
        """)
        translations = {
            'scalar': ':term:`scalar`',
            'array-like': ':py:class:`array-like <numpy.ndarray>`',
        }
        config = Config(
            napoleon_type_aliases=translations, napoleon_preprocess_types=True
        )
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'method')
        assert str(actual) == expected

    def test_raises_types(self):
        docstrings = [
            (
                """
Example Function

Raises
------
  RuntimeError

      A setting wasn't specified, or was invalid.
  ValueError

      Something something value error.

""",
                """
Example Function

:raises RuntimeError: A setting wasn't specified, or was invalid.
:raises ValueError: Something something value error.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
InvalidDimensionsError

""",
                """
Example Function

:raises InvalidDimensionsError:
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
Invalid Dimensions Error

""",
                """
Example Function

:raises Invalid Dimensions Error:
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
Invalid Dimensions Error
    With description

""",
                """
Example Function

:raises Invalid Dimensions Error: With description
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
InvalidDimensionsError
    If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises InvalidDimensionsError: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
Invalid Dimensions Error
    If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises Invalid Dimensions Error: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises If the dimensions couldn't be parsed.:
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
:class:`exc.InvalidDimensionsError`

""",
                """
Example Function

:raises exc.InvalidDimensionsError:
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
:class:`exc.InvalidDimensionsError`
    If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises exc.InvalidDimensionsError: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
:class:`exc.InvalidDimensionsError`
    If the dimensions couldn't be parsed,
    then a :class:`exc.InvalidDimensionsError` will be raised.

""",
                """
Example Function

:raises exc.InvalidDimensionsError: If the dimensions couldn't be parsed,
    then a :class:`exc.InvalidDimensionsError` will be raised.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
:class:`exc.InvalidDimensionsError`
    If the dimensions couldn't be parsed.
:class:`exc.InvalidArgumentsError`
    If the arguments are invalid.

""",
                """
Example Function

:raises exc.InvalidDimensionsError: If the dimensions couldn't be parsed.
:raises exc.InvalidArgumentsError: If the arguments are invalid.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
CustomError
    If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises package.CustomError: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
AnotherError
    If the dimensions couldn't be parsed.

""",
                """
Example Function

:raises ~package.AnotherError: If the dimensions couldn't be parsed.
""",
            ),
            ################################
            (
                """
Example Function

Raises
------
:class:`exc.InvalidDimensionsError`
:class:`exc.InvalidArgumentsError`

""",
                """
Example Function

:raises exc.InvalidDimensionsError:
:raises exc.InvalidArgumentsError:
""",
            ),
        ]
        for docstring, expected in docstrings:
            translations = {
                'CustomError': 'package.CustomError',
                'AnotherError': ':py:exc:`~package.AnotherError`',
            }
            config = Config(
                napoleon_type_aliases=translations, napoleon_preprocess_types=True
            )
            app = mock.Mock()
            actual = NumpyDocstring(docstring, config, app, 'method')
            assert str(actual) == expected

    def test_xrefs_in_return_type(self):
        docstring = """
Example Function

Returns
-------
:class:`numpy.ndarray`
    A :math:`n \\times 2` array containing
    a bunch of math items
"""
        expected = """
Example Function

:returns: A :math:`n \\times 2` array containing
          a bunch of math items
:rtype: :class:`numpy.ndarray`
"""
        config = Config()
        app = mock.Mock()
        actual = NumpyDocstring(docstring, config, app, 'method')
        assert str(actual) == expected

    def test_section_header_underline_length(self):
        docstrings = [
            (
                """
Summary line

Example
-
Multiline example
body

""",
                """
Summary line

Example
-
Multiline example
body
""",
            ),
            ################################
            (
                """
Summary line

Example
--
Multiline example
body

""",
                """
Summary line

.. rubric:: Example

Multiline example
body
""",
            ),
            ################################
            (
                """
Summary line

Example
-------
Multiline example
body

""",
                """
Summary line

.. rubric:: Example

Multiline example
body
""",
            ),
            ################################
            (
                """
Summary line

Example
------------
Multiline example
body

""",
                """
Summary line

.. rubric:: Example

Multiline example
body
""",
            ),
        ]
        for docstring, expected in docstrings:
            actual = NumpyDocstring(docstring)
            assert str(actual) == expected

    def test_list_in_parameter_description(self):
        docstring = """One line summary.

Parameters
----------
no_list : int
one_bullet_empty : int
    *
one_bullet_single_line : int
    - first line
one_bullet_two_lines : int
    +   first line
        continued
two_bullets_single_line : int
    -  first line
    -  second line
two_bullets_two_lines : int
    * first line
      continued
    * second line
      continued
one_enumeration_single_line : int
    1.  first line
one_enumeration_two_lines : int
    1)   first line
         continued
two_enumerations_one_line : int
    (iii) first line
    (iv) second line
two_enumerations_two_lines : int
    a. first line
       continued
    b. second line
       continued
one_definition_one_line : int
    item 1
        first line
one_definition_two_lines : int
    item 1
        first line
        continued
two_definitions_one_line : int
    item 1
        first line
    item 2
        second line
two_definitions_two_lines : int
    item 1
        first line
        continued
    item 2
        second line
        continued
one_definition_blank_line : int
    item 1

        first line

        extra first line

two_definitions_blank_lines : int
    item 1

        first line

        extra first line

    item 2

        second line

        extra second line

definition_after_normal_text : int
    text line

    item 1
        first line
"""

        expected = """One line summary.

:param no_list:
:type no_list: int
:param one_bullet_empty:
                         *
:type one_bullet_empty: int
:param one_bullet_single_line:
                               - first line
:type one_bullet_single_line: int
:param one_bullet_two_lines:
                             +   first line
                                 continued
:type one_bullet_two_lines: int
:param two_bullets_single_line:
                                -  first line
                                -  second line
:type two_bullets_single_line: int
:param two_bullets_two_lines:
                              * first line
                                continued
                              * second line
                                continued
:type two_bullets_two_lines: int
:param one_enumeration_single_line:
                                    1.  first line
:type one_enumeration_single_line: int
:param one_enumeration_two_lines:
                                  1)   first line
                                       continued
:type one_enumeration_two_lines: int
:param two_enumerations_one_line:
                                  (iii) first line
                                  (iv) second line
:type two_enumerations_one_line: int
:param two_enumerations_two_lines:
                                   a. first line
                                      continued
                                   b. second line
                                      continued
:type two_enumerations_two_lines: int
:param one_definition_one_line:
                                item 1
                                    first line
:type one_definition_one_line: int
:param one_definition_two_lines:
                                 item 1
                                     first line
                                     continued
:type one_definition_two_lines: int
:param two_definitions_one_line:
                                 item 1
                                     first line
                                 item 2
                                     second line
:type two_definitions_one_line: int
:param two_definitions_two_lines:
                                  item 1
                                      first line
                                      continued
                                  item 2
                                      second line
                                      continued
:type two_definitions_two_lines: int
:param one_definition_blank_line:
                                  item 1

                                      first line

                                      extra first line
:type one_definition_blank_line: int
:param two_definitions_blank_lines:
                                    item 1

                                        first line

                                        extra first line

                                    item 2

                                        second line

                                        extra second line
:type two_definitions_blank_lines: int
:param definition_after_normal_text: text line

                                     item 1
                                         first line
:type definition_after_normal_text: int
"""
        config = Config(napoleon_use_param=True)
        actual = NumpyDocstring(docstring, config)
        assert str(actual) == expected

        expected = """One line summary.

:Parameters: * **no_list** (:py:class:`int`)
             * **one_bullet_empty** (:py:class:`int`) --

               *
             * **one_bullet_single_line** (:py:class:`int`) --

               - first line
             * **one_bullet_two_lines** (:py:class:`int`) --

               +   first line
                   continued
             * **two_bullets_single_line** (:py:class:`int`) --

               -  first line
               -  second line
             * **two_bullets_two_lines** (:py:class:`int`) --

               * first line
                 continued
               * second line
                 continued
             * **one_enumeration_single_line** (:py:class:`int`) --

               1.  first line
             * **one_enumeration_two_lines** (:py:class:`int`) --

               1)   first line
                    continued
             * **two_enumerations_one_line** (:py:class:`int`) --

               (iii) first line
               (iv) second line
             * **two_enumerations_two_lines** (:py:class:`int`) --

               a. first line
                  continued
               b. second line
                  continued
             * **one_definition_one_line** (:py:class:`int`) --

               item 1
                   first line
             * **one_definition_two_lines** (:py:class:`int`) --

               item 1
                   first line
                   continued
             * **two_definitions_one_line** (:py:class:`int`) --

               item 1
                   first line
               item 2
                   second line
             * **two_definitions_two_lines** (:py:class:`int`) --

               item 1
                   first line
                   continued
               item 2
                   second line
                   continued
             * **one_definition_blank_line** (:py:class:`int`) --

               item 1

                   first line

                   extra first line
             * **two_definitions_blank_lines** (:py:class:`int`) --

               item 1

                   first line

                   extra first line

               item 2

                   second line

                   extra second line
             * **definition_after_normal_text** (:py:class:`int`) -- text line

               item 1
                   first line
"""
        config = Config(napoleon_use_param=False, napoleon_preprocess_types=True)
        actual = NumpyDocstring(docstring, config)
        assert str(actual) == expected

    def test_token_type(self):
        tokens = (
            ('1', 'literal'),
            ('-4.6', 'literal'),
            ('2j', 'literal'),
            ("'string'", 'literal'),
            ('"another_string"', 'literal'),
            ('{1, 2}', 'literal'),
            ("{'va{ue', 'set'}", 'literal'),
            ('optional', 'control'),
            ('default', 'control'),
            (', ', 'delimiter'),
            (' of ', 'delimiter'),
            (' or ', 'delimiter'),
            (': ', 'delimiter'),
            ('True', 'obj'),
            ('None', 'obj'),
            ('name', 'obj'),
            (':py:class:`Enum`', 'reference'),
        )

        for token, expected in tokens:
            actual = _token_type(token)
            assert actual == expected

    def test_tokenize_type_spec(self):
        specs = (
            'str',
            'defaultdict',
            'int, float, or complex',
            'int or float or None, optional',
            'list of list of int or float, optional',
            'tuple of list of str, float, or int',
            '{"F", "C", "N"}',
            "{'F', 'C', 'N'}, default: 'F'",
            "{'F', 'C', 'N or C'}, default 'F'",
            "str, default: 'F or C'",
            'int, default: None',
            'int, default None',
            'int, default :obj:`None`',
            '"ma{icious"',
            r"'with \'quotes\''",
        )

        tokens = (
            ['str'],
            ['defaultdict'],
            ['int', ', ', 'float', ', or ', 'complex'],
            ['int', ' or ', 'float', ' or ', 'None', ', ', 'optional'],
            ['list', ' of ', 'list', ' of ', 'int', ' or ', 'float', ', ', 'optional'],
            ['tuple', ' of ', 'list', ' of ', 'str', ', ', 'float', ', or ', 'int'],
            ['{', '"F"', ', ', '"C"', ', ', '"N"', '}'],
            ['{', "'F'", ', ', "'C'", ', ', "'N'", '}', ', ', 'default', ': ', "'F'"],
            ['{', "'F'", ', ', "'C'", ', ', "'N or C'", '}', ', ', 'default', ' ', "'F'"],
            ['str', ', ', 'default', ': ', "'F or C'"],
            ['int', ', ', 'default', ': ', 'None'],
            ['int', ', ', 'default', ' ', 'None'],
            ['int', ', ', 'default', ' ', ':obj:`None`'],
            ['"ma{icious"'],
            [r"'with \'quotes\''"],
        )  # fmt: skip

        for spec, expected in zip(specs, tokens, strict=True):
            actual = _tokenize_type_spec(spec)
            assert actual == expected

    def test_recombine_set_tokens(self):
        tokens = (
            ['{', '1', ', ', '2', '}'],
            ['{', '"F"', ', ', '"C"', ', ', '"N"', '}', ', ', 'optional'],
            ['{', "'F'", ', ', "'C'", ', ', "'N'", '}', ', ', 'default', ': ', 'None'],
            ['{', "'F'", ', ', "'C'", ', ', "'N'", '}', ', ', 'default', ' ', 'None'],
        )

        combined_tokens = (
            ['{1, 2}'],
            ['{"F", "C", "N"}', ', ', 'optional'],
            ["{'F', 'C', 'N'}", ', ', 'default', ': ', 'None'],
            ["{'F', 'C', 'N'}", ', ', 'default', ' ', 'None'],
        )

        for tokens_, expected in zip(tokens, combined_tokens, strict=True):
            actual = _recombine_set_tokens(tokens_)
            assert actual == expected

    def test_recombine_set_tokens_invalid(self):
        tokens = (
            ['{', '1', ', ', '2'],
            ['"F"', ', ', '"C"', ', ', '"N"', '}', ', ', 'optional'],
            ['{', '1', ', ', '2', ', ', 'default', ': ', 'None'],
        )
        combined_tokens = (
            ['{1, 2'],
            ['"F"', ', ', '"C"', ', ', '"N"', '}', ', ', 'optional'],
            ['{1, 2', ', ', 'default', ': ', 'None'],
        )

        for tokens_, expected in zip(tokens, combined_tokens, strict=True):
            actual = _recombine_set_tokens(tokens_)
            assert actual == expected

    def test_convert_numpy_type_spec(self):
        translations = {
            'DataFrame': 'pandas.DataFrame',
        }

        specs = (
            '',
            'optional',
            'str, optional',
            'int or float or None, default: None',
            'list of tuple of str, optional',
            'int, default None',
            '{"F", "C", "N"}',
            "{'F', 'C', 'N'}, default: 'N'",
            "{'F', 'C', 'N'}, default 'N'",
            'DataFrame, optional',
        )

        converted = (
            '',
            '*optional*',
            ':py:class:`str`, *optional*',
            ':py:class:`int` or :py:class:`float` or :py:obj:`None`, *default*: :py:obj:`None`',
            ':py:class:`list` of :py:class:`tuple` of :py:class:`str`, *optional*',
            ':py:class:`int`, *default* :py:obj:`None`',
            '``{"F", "C", "N"}``',
            "``{'F', 'C', 'N'}``, *default*: ``'N'``",
            "``{'F', 'C', 'N'}``, *default* ``'N'``",
            ':py:class:`pandas.DataFrame`, *optional*',
        )

        for spec, expected in zip(specs, converted, strict=True):
            actual = _convert_type_spec(spec, translations=translations)
            assert actual == expected

    def test_parameter_types(self):
        docstring = dedent("""\
            Parameters
            ----------
            param1 : DataFrame
                the data to work on
            param2 : int or float or None, optional
                a parameter with different types
            param3 : dict-like, optional
                a optional mapping
            param4 : int or float or None, optional
                a optional parameter with different types
            param5 : {"F", "C", "N"}, optional
                a optional parameter with fixed values
            param6 : int, default None
                different default format
            param7 : mapping of hashable to str, optional
                a optional mapping
            param8 : ... or Ellipsis
                ellipsis
            param9 : tuple of list of int
                a parameter with tuple of list of int
        """)
        expected = dedent("""\
            :param param1: the data to work on
            :type param1: :py:class:`DataFrame`
            :param param2: a parameter with different types
            :type param2: :py:class:`int` or :py:class:`float` or :py:obj:`None`, *optional*
            :param param3: a optional mapping
            :type param3: :term:`dict-like <mapping>`, *optional*
            :param param4: a optional parameter with different types
            :type param4: :py:class:`int` or :py:class:`float` or :py:obj:`None`, *optional*
            :param param5: a optional parameter with fixed values
            :type param5: ``{"F", "C", "N"}``, *optional*
            :param param6: different default format
            :type param6: :py:class:`int`, *default* :py:obj:`None`
            :param param7: a optional mapping
            :type param7: :term:`mapping` of :term:`hashable` to :py:class:`str`, *optional*
            :param param8: ellipsis
            :type param8: :py:obj:`... <Ellipsis>` or :py:obj:`Ellipsis`
            :param param9: a parameter with tuple of list of int
            :type param9: :py:class:`tuple` of :py:class:`list` of :py:class:`int`
        """)
        translations = {
            'dict-like': ':term:`dict-like <mapping>`',
            'mapping': ':term:`mapping`',
            'hashable': ':term:`hashable`',
        }
        config = Config(
            napoleon_use_param=True,
            napoleon_use_rtype=True,
            napoleon_preprocess_types=True,
            napoleon_type_aliases=translations,
        )
        actual = NumpyDocstring(docstring, config)
        assert str(actual) == expected

    @pytest.mark.sphinx('html', testroot='root')
    def test_token_type_invalid(self, app):
        tokens = (
            '{1, 2',
            '}',
            "'abc",
            "def'",
            '"ghi',
            'jkl"',
        )
        errors = (
            r'.+: invalid value set \(missing closing brace\):',
            r'.+: invalid value set \(missing opening brace\):',
            r'.+: malformed string literal \(missing closing quote\):',
            r'.+: malformed string literal \(missing opening quote\):',
            r'.+: malformed string literal \(missing closing quote\):',
            r'.+: malformed string literal \(missing opening quote\):',
        )
        for token, error in zip(tokens, errors, strict=True):
            try:
                _token_type(token)
            finally:
                raw_warnings = app.warning.getvalue()
                warnings = [w for w in raw_warnings.split('\n') if w.strip()]

                assert len(warnings) == 1
                assert re.compile(error).match(warnings[0])
                app.warning.truncate(0)

    @pytest.mark.parametrize(
        ('name', 'expected'),
        [
            ('x, y, z', 'x, y, z'),
            ('*args, **kwargs', r'\*args, \*\*kwargs'),
            ('*x, **y', r'\*x, \*\*y'),
        ],
    )
    def test_escape_args_and_kwargs(self, name, expected):
        numpy_docstring = NumpyDocstring('')
        actual = numpy_docstring._escape_args_and_kwargs(name)

        assert actual == expected

    def test_pep526_annotations(self):
        # test class attributes annotations
        config = Config(
            napoleon_attr_annotations=True,
        )
        actual = NumpyDocstring(
            cleandoc(PEP526NumpyClass.__doc__),
            config,
            app=None,
            what='class',
            obj=PEP526NumpyClass,
        )
        expected = """\
Sample class with PEP 526 annotations and numpy docstring

.. attribute:: attr1

   Attr1 description

   :type: int

.. attribute:: attr2

   Attr2 description

   :type: str
"""
        print(actual)
        assert str(actual) == expected


@pytest.mark.sphinx(
    'text',
    testroot='ext-napoleon',
    confoverrides={
        'autodoc_typehints': 'description',
        'autodoc_typehints_description_target': 'all',
    },
)
def test_napoleon_and_autodoc_typehints_description_all(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'typehints.txt').read_text(encoding='utf-8')
    assert content == (
        'typehints\n'
        '*********\n'
        '\n'
        'mypackage.typehints.hello(x, *args, **kwargs)\n'
        '\n'
        '   Parameters:\n'
        '      * **x** (*int*) -- X\n'
        '\n'
        '      * ***args** (*int*) -- Additional arguments.\n'
        '\n'
        '      * ****kwargs** (*int*) -- Extra arguments.\n'
        '\n'
        '   Return type:\n'
        '      None\n'
    )


@pytest.mark.sphinx(
    'text',
    testroot='ext-napoleon',
    confoverrides={
        'autodoc_typehints': 'description',
        'autodoc_typehints_description_target': 'documented_params',
    },
)
def test_napoleon_and_autodoc_typehints_description_documented_params(
    app: SphinxTestApp,
) -> None:
    app.build()
    content = (app.outdir / 'typehints.txt').read_text(encoding='utf-8')
    assert content == (
        'typehints\n'
        '*********\n'
        '\n'
        'mypackage.typehints.hello(x, *args, **kwargs)\n'
        '\n'
        '   Parameters:\n'
        '      * **x** (*int*) -- X\n'
        '\n'
        '      * ***args** (*int*) -- Additional arguments.\n'
        '\n'
        '      * ****kwargs** (*int*) -- Extra arguments.\n'
    )


@pytest.mark.sphinx('html', testroot='ext-napoleon-paramtype', freshenv=True)
def test_napoleon_keyword_and_paramtype(app, tmp_path):
    inv_file = tmp_path / 'objects.inv'
    inv_file.write_bytes(
        b"""\
# Sphinx inventory version 2
# Project: Intersphinx Test
# Version: 42
# The remainder of this file is compressed using zlib.
"""
        + zlib.compress(b"""\
None py:data 1 none.html -
list py:class 1 list.html -
int py:class 1 int.html -
""")
    )
    app.config.intersphinx_mapping = {'python': ('127.0.0.1:5555', str(inv_file))}
    validate_intersphinx_mapping(app, app.config)
    load_mappings(app)

    app.build(force_all=True)

    etree = etree_parse(app.outdir / 'index.html')

    for name, typename in product(
        ('keyword', 'kwarg', 'kwparam'),
        ('paramtype', 'kwtype'),
    ):
        param = f'{name}_{typename}'
        li_ = list(etree.findall(f'.//li/p/strong[.="{param}"]/../..'))
        assert len(li_) == 1
        li = li_[0]

        text = li.text or ''.join(li.itertext())
        assert text == f'{param} (list[int]) \u2013 some param'

        a_ = list(li.findall('.//a[@class="reference external"]'))

        assert len(a_) == 2
        for a, uri in zip(a_, ('list.html', 'int.html'), strict=True):
            assert a.attrib['href'] == f'127.0.0.1:5555/{uri}'
            assert a.attrib['title'] == '(in Intersphinx Test v42)'



"""Tests sphinx.util.matching functions."""

from __future__ import annotations

from sphinx.util.matching import Matcher, compile_matchers, get_matching_files


def test_compile_matchers() -> None:
    # exact matching
    pat = compile_matchers(['hello.py']).pop()
    assert pat('hello.py')
    assert not pat('hello-py')
    assert not pat('subdir/hello.py')

    # wild card (*)
    pat = compile_matchers(['hello.*']).pop()
    assert pat('hello.py')
    assert pat('hello.rst')

    pat = compile_matchers(['*.py']).pop()
    assert pat('hello.py')
    assert pat('world.py')
    assert not pat('subdir/hello.py')

    # wild card (**)
    pat = compile_matchers(['hello.**']).pop()
    assert pat('hello.py')
    assert pat('hello.rst')
    assert pat('hello.py/world.py')

    pat = compile_matchers(['**.py']).pop()
    assert pat('hello.py')
    assert pat('world.py')
    assert pat('subdir/hello.py')

    pat = compile_matchers(['**/hello.py']).pop()
    assert not pat('hello.py')
    assert pat('subdir/hello.py')
    assert pat('subdir/subdir/hello.py')

    # wild card (?)
    pat = compile_matchers(['hello.?']).pop()
    assert pat('hello.c')
    assert not pat('hello.py')

    # pattern ([...])
    pat = compile_matchers(['hello[12\\].py']).pop()
    assert pat('hello1.py')
    assert pat('hello2.py')
    assert pat('hello\\.py')
    assert not pat('hello3.py')

    pat = compile_matchers(['hello[^12].py']).pop()  # "^" is not negative identifier
    assert pat('hello1.py')
    assert pat('hello2.py')
    assert pat('hello^.py')
    assert not pat('hello3.py')

    # negative pattern ([!...])
    pat = compile_matchers(['hello[!12].py']).pop()
    assert not pat('hello1.py')
    assert not pat('hello2.py')
    assert not pat('hello/.py')  # negative pattern does not match to "/"
    assert pat('hello3.py')

    # non patterns
    pat = compile_matchers(['hello[.py']).pop()
    assert pat('hello[.py')
    assert not pat('hello.py')

    pat = compile_matchers(['hello[].py']).pop()
    assert pat('hello[].py')
    assert not pat('hello.py')

    pat = compile_matchers(['hello[!].py']).pop()
    assert pat('hello[!].py')
    assert not pat('hello.py')


def test_Matcher() -> None:
    matcher = Matcher(['hello.py', '**/world.py'])
    assert matcher('hello.py')
    assert not matcher('subdir/hello.py')
    assert matcher('world.py')
    assert matcher('subdir/world.py')


def test_get_matching_files_all(rootdir):
    files = get_matching_files(rootdir / 'test-root')
    assert sorted(files) == [
        'Makefile',
        '_templates/contentssb.html',
        '_templates/customsb.html',
        '_templates/layout.html',
        'autodoc.txt',
        'autodoc_target.py',
        'bom.txt',
        'conf.py',
        'extapi.txt',
        'extensions.txt',
        'file_with_special_#_chars.xyz',
        'footnote.txt',
        'images.txt',
        'img.foo.png',
        'img.gif',
        'img.pdf',
        'img.png',
        'includes.txt',
        'index.txt',
        'lists.txt',
        'literal.inc',
        'literal_orig.inc',
        'markup.txt',
        'math.txt',
        'objects.txt',
        'otherext.foo',
        'parsermod.py',
        'quotes.inc',
        'rimg.png',
        'special/api.h',
        'special/code.py',
        'subdir/excluded.txt',
        'subdir/images.txt',
        'subdir/img.png',
        'subdir/include.inc',
        'subdir/includes.txt',
        'subdir/simg.png',
        'svgimg.pdf',
        'svgimg.svg',
        'tabs.inc',
        'test.inc',
        'wrongenc.inc',
    ]


def test_get_matching_files_all_exclude_single(rootdir):
    files = get_matching_files(rootdir / 'test-root', exclude_patterns=['**.html'])
    assert sorted(files) == [
        'Makefile',
        'autodoc.txt',
        'autodoc_target.py',
        'bom.txt',
        'conf.py',
        'extapi.txt',
        'extensions.txt',
        'file_with_special_#_chars.xyz',
        'footnote.txt',
        'images.txt',
        'img.foo.png',
        'img.gif',
        'img.pdf',
        'img.png',
        'includes.txt',
        'index.txt',
        'lists.txt',
        'literal.inc',
        'literal_orig.inc',
        'markup.txt',
        'math.txt',
        'objects.txt',
        'otherext.foo',
        'parsermod.py',
        'quotes.inc',
        'rimg.png',
        'special/api.h',
        'special/code.py',
        'subdir/excluded.txt',
        'subdir/images.txt',
        'subdir/img.png',
        'subdir/include.inc',
        'subdir/includes.txt',
        'subdir/simg.png',
        'svgimg.pdf',
        'svgimg.svg',
        'tabs.inc',
        'test.inc',
        'wrongenc.inc',
    ]


def test_get_matching_files_all_exclude_multiple(rootdir):
    files = get_matching_files(
        rootdir / 'test-root', exclude_patterns=['**.html', '**.inc']
    )
    assert sorted(files) == [
        'Makefile',
        'autodoc.txt',
        'autodoc_target.py',
        'bom.txt',
        'conf.py',
        'extapi.txt',
        'extensions.txt',
        'file_with_special_#_chars.xyz',
        'footnote.txt',
        'images.txt',
        'img.foo.png',
        'img.gif',
        'img.pdf',
        'img.png',
        'includes.txt',
        'index.txt',
        'lists.txt',
        'markup.txt',
        'math.txt',
        'objects.txt',
        'otherext.foo',
        'parsermod.py',
        'rimg.png',
        'special/api.h',
        'special/code.py',
        'subdir/excluded.txt',
        'subdir/images.txt',
        'subdir/img.png',
        'subdir/includes.txt',
        'subdir/simg.png',
        'svgimg.pdf',
        'svgimg.svg',
    ]


def test_get_matching_files_all_exclude_nonexistent(rootdir):
    files = get_matching_files(rootdir / 'test-root', exclude_patterns=['halibut/**'])
    assert sorted(files) == [
        'Makefile',
        '_templates/contentssb.html',
        '_templates/customsb.html',
        '_templates/layout.html',
        'autodoc.txt',
        'autodoc_target.py',
        'bom.txt',
        'conf.py',
        'extapi.txt',
        'extensions.txt',
        'file_with_special_#_chars.xyz',
        'footnote.txt',
        'images.txt',
        'img.foo.png',
        'img.gif',
        'img.pdf',
        'img.png',
        'includes.txt',
        'index.txt',
        'lists.txt',
        'literal.inc',
        'literal_orig.inc',
        'markup.txt',
        'math.txt',
        'objects.txt',
        'otherext.foo',
        'parsermod.py',
        'quotes.inc',
        'rimg.png',
        'special/api.h',
        'special/code.py',
        'subdir/excluded.txt',
        'subdir/images.txt',
        'subdir/img.png',
        'subdir/include.inc',
        'subdir/includes.txt',
        'subdir/simg.png',
        'svgimg.pdf',
        'svgimg.svg',
        'tabs.inc',
        'test.inc',
        'wrongenc.inc',
    ]


def test_get_matching_files_all_include_single(rootdir):
    files = get_matching_files(rootdir / 'test-root', include_patterns=['subdir/**'])
    assert sorted(files) == [
        'subdir/excluded.txt',
        'subdir/images.txt',
        'subdir/img.png',
        'subdir/include.inc',
        'subdir/includes.txt',
        'subdir/simg.png',
    ]


def test_get_matching_files_all_include_multiple(rootdir):
    files = get_matching_files(
        rootdir / 'test-root', include_patterns=['special/**', 'subdir/**']
    )
    assert sorted(files) == [
        'special/api.h',
        'special/code.py',
        'subdir/excluded.txt',
        'subdir/images.txt',
        'subdir/img.png',
        'subdir/include.inc',
        'subdir/includes.txt',
        'subdir/simg.png',
    ]


def test_get_matching_files_all_include_nonexistent(rootdir):
    files = get_matching_files(rootdir / 'test-root', include_patterns=['halibut/**'])
    assert sorted(files) == []


def test_get_matching_files_all_include_prefix(rootdir):
    files = get_matching_files(rootdir / 'test-root', include_patterns=['autodoc*'])
    assert sorted(files) == [
        'autodoc.txt',
        'autodoc_target.py',
    ]


def test_get_matching_files_all_include_question_mark(rootdir):
    files = get_matching_files(rootdir / 'test-root', include_patterns=['img.???'])
    assert sorted(files) == [
        'img.gif',
        'img.pdf',
        'img.png',
    ]



"""Tests the transformations"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='_blank')
def test_transforms_reorder_consecutive_target_and_index_nodes_preserve_order(
    app: SphinxTestApp,
) -> None:
    text = '.. index:: abc\n.. index:: def\n.. index:: ghi\n.. index:: jkl\n\ntext\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            nodes.target,
            nodes.target,
            nodes.paragraph,
        ),
    )
    assert_node(
        doctree[0], addnodes.index, entries=[('single', 'abc', 'index-0', '', None)]
    )
    assert_node(
        doctree[1], addnodes.index, entries=[('single', 'def', 'index-1', '', None)]
    )
    assert_node(
        doctree[2], addnodes.index, entries=[('single', 'ghi', 'index-2', '', None)]
    )
    assert_node(
        doctree[3], addnodes.index, entries=[('single', 'jkl', 'index-3', '', None)]
    )
    assert_node(doctree[4], nodes.target, refid='index-0')
    assert_node(doctree[5], nodes.target, refid='index-1')
    assert_node(doctree[6], nodes.target, refid='index-2')
    assert_node(doctree[7], nodes.target, refid='index-3')
    # assert_node(doctree[8], nodes.paragraph)


@pytest.mark.sphinx('html', testroot='_blank')
def test_transforms_reorder_consecutive_target_and_index_nodes_no_merge_across_other_nodes(
    app,
):
    text = (
        '.. index:: abc\n'
        '.. index:: def\n'
        '\n'
        'text\n'
        '\n'
        '.. index:: ghi\n'
        '.. index:: jkl\n'
        '\n'
        'text\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            nodes.paragraph,
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            nodes.paragraph,
        ),
    )
    assert_node(
        doctree[0], addnodes.index, entries=[('single', 'abc', 'index-0', '', None)]
    )
    assert_node(
        doctree[1], addnodes.index, entries=[('single', 'def', 'index-1', '', None)]
    )
    assert_node(doctree[2], nodes.target, refid='index-0')
    assert_node(doctree[3], nodes.target, refid='index-1')
    # assert_node(doctree[4], nodes.paragraph)
    assert_node(
        doctree[5], addnodes.index, entries=[('single', 'ghi', 'index-2', '', None)]
    )
    assert_node(
        doctree[6], addnodes.index, entries=[('single', 'jkl', 'index-3', '', None)]
    )
    assert_node(doctree[7], nodes.target, refid='index-2')
    assert_node(doctree[8], nodes.target, refid='index-3')
    # assert_node(doctree[9], nodes.paragraph)


@pytest.mark.sphinx('html', testroot='_blank')
def test_transforms_reorder_consecutive_target_and_index_nodes_merge_with_labels(
    app: SphinxTestApp,
) -> None:
    text = (
        '.. _abc:\n'
        '.. index:: def\n'
        '.. _ghi:\n'
        '.. index:: jkl\n'
        '.. _mno:\n'
        '\n'
        'Heading\n'
        '=======\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            nodes.title,
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            nodes.target,
            nodes.target,
            nodes.target,
        ),
    )
    # assert_node(doctree[8], nodes.title)
    assert_node(
        doctree[1], addnodes.index, entries=[('single', 'def', 'index-0', '', None)]
    )
    assert_node(
        doctree[2], addnodes.index, entries=[('single', 'jkl', 'index-1', '', None)]
    )
    assert_node(doctree[3], nodes.target, refid='abc')
    assert_node(doctree[4], nodes.target, refid='index-0')
    assert_node(doctree[5], nodes.target, refid='ghi')
    assert_node(doctree[6], nodes.target, refid='index-1')
    assert_node(doctree[7], nodes.target, refid='mno')



"""Tests the reStructuredText domain."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_signature,
)
from sphinx.domains.rst import parse_directive
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


def test_parse_directive() -> None:
    s = parse_directive(' foö  ')
    assert s == ('foö', '')

    s = parse_directive(' ..    foö ::  ')
    assert s == ('foö', '')

    s = parse_directive('.. foö:: args1 args2')
    assert s == ('foö', ' args1 args2')

    s = parse_directive('.. :: bar')
    assert s == ('.. :: bar', '')


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_directive(app: SphinxTestApp) -> None:
    # bare
    text = '.. rst:directive:: toctree'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [desc, ([desc_signature, desc_name, '.. toctree::'], [desc_content, ()])],
        ),
    )
    assert_node(
        doctree[0],
        entries=[('single', 'toctree (directive)', 'directive-toctree', '', None)],
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='directive',
        domain='rst',
        objtype='directive',
        no_index=False,
    )

    # decorated
    text = '.. rst:directive:: .. toctree::'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [desc, ([desc_signature, desc_name, '.. toctree::'], [desc_content, ()])],
        ),
    )
    assert_node(
        doctree[0],
        entries=[('single', 'toctree (directive)', 'directive-toctree', '', None)],
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='directive',
        domain='rst',
        objtype='directive',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_directive_with_argument(app: SphinxTestApp) -> None:
    text = '.. rst:directive:: .. toctree:: foo bar baz'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        ([desc_name, '.. toctree::'], [desc_addname, ' foo bar baz']),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        entries=[('single', 'toctree (directive)', 'directive-toctree', '', None)],
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='directive',
        domain='rst',
        objtype='directive',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_directive_option(app: SphinxTestApp) -> None:
    text = '.. rst:directive:option:: foo'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [desc, ([desc_signature, desc_name, ':foo:'], [desc_content, ()])],
        ),
    )
    assert_node(
        doctree[0],
        entries=[
            ('single', ':foo: (directive option)', 'directive-option-foo', '', 'F')
        ],
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='directive:option',
        domain='rst',
        objtype='directive:option',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_directive_option_with_argument(app: SphinxTestApp) -> None:
    text = '.. rst:directive:option:: foo: bar baz'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        ([desc_name, ':foo:'], [desc_annotation, ' bar baz']),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        entries=[
            ('single', ':foo: (directive option)', 'directive-option-foo', '', 'F')
        ],
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='directive:option',
        domain='rst',
        objtype='directive:option',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_directive_option_type(app: SphinxTestApp) -> None:
    text = '.. rst:directive:option:: foo\n   :type: directives.flags\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ':foo:'],
                            [desc_annotation, ' (directives.flags)'],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        entries=[
            ('single', ':foo: (directive option)', 'directive-option-foo', '', 'F')
        ],
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='directive:option',
        domain='rst',
        objtype='directive:option',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_directive_and_directive_option(app):
    text = '.. rst:directive:: foo\n\n   .. rst:directive:option:: bar\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, desc_name, '.. foo::'],
                    [desc_content, (addnodes.index, desc)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[1][1][0],
        entries=[
            (
                'pair',
                'foo (directive); :bar: (directive option)',
                'directive-option-foo-bar',
                '',
                'B',
            )
        ],
    )
    assert_node(
        doctree[1][1][1], ([desc_signature, desc_name, ':bar:'], [desc_content, ()])
    )
    assert_node(
        doctree[1][1][1],
        addnodes.desc,
        desctype='directive:option',
        domain='rst',
        objtype='directive:option',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_rst_role(app: SphinxTestApp) -> None:
    text = '.. rst:role:: ref'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [desc, ([desc_signature, desc_name, ':ref:'], [desc_content, ()])],
        ),
    )
    assert_node(doctree[0], entries=[('single', 'ref (role)', 'role-ref', '', None)])
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='role',
        domain='rst',
        objtype='role',
        no_index=False,
    )



from __future__ import annotations

from sphinx.util._uri import encode_uri


def test_encode_uri() -> None:
    expected = (
        'https://ru.wikipedia.org/wiki/%D0%A1%D0%B8%D1%81%D1%82%D0%B5%D0%BC%D0%B0_'
        '%D1%83%D0%BF%D1%80%D0%B0%D0%B2%D0%BB%D0%B5%D0%BD%D0%B8%D1%8F_'
        '%D0%B1%D0%B0%D0%B7%D0%B0%D0%BC%D0%B8_%D0%B4%D0%B0%D0%BD%D0%BD%D1%8B%D1%85'
    )
    uri = 'https://ru.wikipedia.org/wiki/Система_управления_базами_данных'
    assert encode_uri(uri) == expected

    expected = (
        'https://github.com/search?utf8=%E2%9C%93&q=is%3Aissue+is%3Aopen+is%3A'
        'sprint-friendly+user%3Ajupyter&type=Issues&ref=searchresults'
    )
    uri = (
        'https://github.com/search?utf8=✓&q=is%3Aissue+is%3Aopen+is%3A'
        'sprint-friendly+user%3Ajupyter&type=Issues&ref=searchresults'
    )
    assert encode_uri(uri) == expected



"""Tests the directives"""

from __future__ import annotations

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

DOMAINS = [
    # directive, no-index, no-index-entry, signature of f, signature of g, index entry of g
    (
        'c:function',
        False,
        True,
        'void f()',
        'void g()',
        ('single', 'g (C function)', 'c.g', '', None),
    ),
    (
        'cpp:function',
        False,
        True,
        'void f()',
        'void g()',
        ('single', 'g (C++ function)', '_CPPv41gv', '', None),
    ),
    (
        'js:function',
        True,
        True,
        'f()',
        'g()',
        ('single', 'g() (built-in function)', 'g', '', None),
    ),
    (
        'py:function',
        True,
        True,
        'f()',
        'g()',
        ('pair', 'built-in function; g()', 'g', '', None),
    ),
    (
        'rst:directive',
        True,
        False,
        'f',
        'g',
        ('single', 'g (directive)', 'directive-g', '', None),
    ),
    (
        'cmdoption',
        True,
        False,
        'f',
        'g',
        ('pair', 'command line option; g', 'cmdoption-arg-g', '', None),
    ),
    (
        'envvar',
        True,
        False,
        'f',
        'g',
        ('single', 'environment variable; g', 'envvar-g', '', None),
    ),
]


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    text = f'.. {directive}:: {sig_f}\n   :no-typesetting:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, nodes.target))


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting_twice(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    text = (
        f'.. {directive}:: {sig_f}\n'
        f'   :no-typesetting:\n'
        f'.. {directive}:: {sig_g}\n'
        f'   :no-typesetting:\n'
    )
    doctree = restructuredtext.parse(app, text)
    # Note that all index nodes come before the target nodes
    assert_node(doctree, (addnodes.index, addnodes.index, nodes.target, nodes.target))


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting_noindex_orig(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    if not no_index:
        pytest.skip(f'{directive} does not support :no-index: option')
    text = f'.. {directive}:: {sig_f}\n   :no-index:\n.. {directive}:: {sig_g}\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, addnodes.desc, addnodes.index, addnodes.desc))
    assert_node(doctree[2], addnodes.index, entries=[index_g])


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting_noindex(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    if not no_index:
        pytest.skip(f'{directive} does not support :no-index: option')
    text = (
        f'.. {directive}:: {sig_f}\n'
        f'   :no-index:\n'
        f'   :no-typesetting:\n'
        f'.. {directive}:: {sig_g}\n'
        f'   :no-typesetting:\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, addnodes.index, nodes.target))
    assert_node(doctree[0], addnodes.index, entries=[])
    assert_node(doctree[1], addnodes.index, entries=[index_g])


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting_no_index_entry(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    if not no_index_entry:
        pytest.skip(f'{directive} does not support :no-index-entry: option')
    text = (
        f'.. {directive}:: {sig_f}\n'
        f'   :no-index-entry:\n'
        f'   :no-typesetting:\n'
        f'.. {directive}:: {sig_g}\n'
        f'   :no-typesetting:\n'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, addnodes.index, nodes.target, nodes.target))
    assert_node(doctree[0], addnodes.index, entries=[])
    assert_node(doctree[1], addnodes.index, entries=[index_g])


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting_code(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    text = (
        f'.. {directive}:: {sig_f}\n'
        f'   :no-typesetting:\n'
        f'.. {directive}:: {sig_g}\n'
        f'   :no-typesetting:\n'
        f'.. code::\n'
        f'\n'
        f'   code\n'
    )
    doctree = restructuredtext.parse(app, text)
    # Note that all index nodes come before the targets
    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            nodes.literal_block,
        ),
    )


@pytest.mark.parametrize(
    ('directive', 'no_index', 'no_index_entry', 'sig_f', 'sig_g', 'index_g'),
    DOMAINS,
)
@pytest.mark.sphinx('html', testroot='root')
def test_object_description_no_typesetting_heading(
    app, directive, no_index, no_index_entry, sig_f, sig_g, index_g
):
    text = (
        f'.. {directive}:: {sig_f}\n'
        f'   :no-typesetting:\n'
        f'.. {directive}:: {sig_g}\n'
        f'   :no-typesetting:\n'
        f'\n'
        f'Heading\n'
        f'=======\n'
    )
    doctree = restructuredtext.parse(app, text)
    # Note that all index nodes come before the targets and the heading is floated before those.
    assert_node(
        doctree,
        (nodes.title, addnodes.index, addnodes.index, nodes.target, nodes.target),
    )



"""module with trailing underscores everywhere"""


class class_:
    """Class"""

    def method_(_arg):  # NoQA: N805
        """Method"""
        pass


def function_(_arg):
    """Function"""
    pass



latex_documents = [
    (
        'index',
        'test.tex',
        'The basic Sphinx documentation for testing',
        'Sphinx',
        'report',
    )
]
html_logo = 'https://www.python.org/static/img/python-logo.png'
html_favicon = 'https://www.python.org/static/favicon.ico'



"""Test pycode.parser."""

from __future__ import annotations

from sphinx.pycode.parser import Parser
from sphinx.util.inspect import signature_from_str


def test_comment_picker_basic() -> None:
    source = (
        'a = 1 + 1      #: assignment\n'
        'b = 1 +\\\n 1  #: assignment including a CR\n'
        'c = (1 +\n 1)  #: tuple  \n'
        'd = {1, \n 1}  #:     set\n'
        'e = [1, \n 1]  #: list #: additional comment\n'
        'f = "abc"\n'
        '#: string; comment on next line (ignored)\n'
        'g = 1.0\n'
        '"""float; string on next line"""\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('', 'a'): 'assignment',
        ('', 'b'): 'assignment including a CR',
        ('', 'c'): 'tuple  ',
        ('', 'd'): '    set',
        ('', 'e'): 'list #: additional comment',
        ('', 'g'): 'float; string on next line',
    }


def test_comment_picker_location() -> None:
    # multiple "before" comments
    source = (
        '#: comment before assignment1\n#:\n#: comment before assignment2\na = 1 + 1\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('', 'a'): 'comment before assignment1\n\ncomment before assignment2'
    }

    # before and after comments
    source = '#: comment before assignment\na = 1 + 1  #: comment after assignment\n'
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'a'): 'comment after assignment'}

    # after comment and next line string
    source = 'a = 1 + 1\n  #: comment after assignment\n"""string on next line"""\n'
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'a'): 'string on next line'}

    # before comment and next line string
    source = '#: comment before assignment\na = 1 + 1\n"""string on next line"""\n'
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'a'): 'string on next line'}

    # before comment, after comment and next line string
    source = (
        '#: comment before assignment\n'
        'a = 1 + 1  #: comment after assignment\n'
        '"""string on next line"""\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'a'): 'string on next line'}

    # inside __init__ method
    source = (
        'class Foo(object):\n'
        '    def __init__(self):\n'
        '        #: comment before assignment\n'
        '        self.attr1 = None\n'
        '        self.attr2 = None  #: comment after assignment\n'
        '\n'
        '        #: comment for attr3(1)\n'
        '        self.attr3 = None  #: comment for attr3(2)\n'
        '        """comment for attr3(3)"""\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('Foo', 'attr1'): 'comment before assignment',
        ('Foo', 'attr2'): 'comment after assignment',
        ('Foo', 'attr3'): 'comment for attr3(3)',
    }


def test_annotated_assignment() -> None:
    source = (
        'a: str = "Sphinx"  #: comment\n'
        'b: int = 1\n'
        '"""string on next line"""\n'
        'c: int  #: comment\n'
        'd = 1  # type: int\n'
        '"""string on next line"""\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('', 'a'): 'comment',
        ('', 'b'): 'string on next line',
        ('', 'c'): 'comment',
        ('', 'd'): 'string on next line',
    }
    assert parser.annotations == {
        ('', 'a'): 'str',
        ('', 'b'): 'int',
        ('', 'c'): 'int',
        ('', 'd'): 'int',
    }
    assert parser.definitions == {}


def test_complex_assignment() -> None:
    source = (
        'a = 1 + 1; b = a  #: compound statement\n'
        'c, d = (1, 1)  #: unpack assignment\n'
        'e = True  #: first assignment\n'
        'e = False  #: second assignment\n'
        'f = g = None  #: multiple assignment at once\n'
        '(theta, phi) = (0, 0.5)  #: unpack assignment via tuple\n'
        '[x, y] = (5, 6)  #: unpack assignment via list\n'
        'h, *i, j = (1, 2, 3, 4)  #: unpack assignment2\n'
        'k, *self.attr = (5, 6, 7)  #: unpack assignment3\n'
        'l, *m[0] = (8, 9, 0)  #: unpack assignment4\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('', 'b'): 'compound statement',
        ('', 'c'): 'unpack assignment',
        ('', 'd'): 'unpack assignment',
        ('', 'e'): 'second assignment',
        ('', 'f'): 'multiple assignment at once',
        ('', 'g'): 'multiple assignment at once',
        ('', 'theta'): 'unpack assignment via tuple',
        ('', 'phi'): 'unpack assignment via tuple',
        ('', 'x'): 'unpack assignment via list',
        ('', 'y'): 'unpack assignment via list',
        ('', 'h'): 'unpack assignment2',
        ('', 'i'): 'unpack assignment2',
        ('', 'j'): 'unpack assignment2',
        ('', 'k'): 'unpack assignment3',
        ('', 'l'): 'unpack assignment4',
    }
    assert parser.definitions == {}


def test_assignment_in_try_clause() -> None:
    source = (
        'try:\n'
        '    a = None  #: comment\n'
        'except:\n'
        '    b = None  #: ignored\n'
        'else:\n'
        '    c = None  #: comment\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'a'): 'comment', ('', 'c'): 'comment'}
    assert parser.deforders == {'a': 0, 'c': 1}


def test_obj_assignment() -> None:
    source = (
        'obj = SomeObject()  #: some object\n'
        'obj.attr = 1  #: attr1\n'
        'obj.attr.attr = 1  #: attr2\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'obj'): 'some object'}
    assert parser.definitions == {}


def test_container_assignment() -> None:
    source = (
        'l = []  #: list\n'
        'l[1] = True  #: list assignment\n'
        'l[0:0] = []  #: list assignment\n'
        'l[_from:_to] = []  #: list assignment\n'
        'd = {}  #: dict\n'
        'd["doc"] = 1  #: dict assignment\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('', 'l'): 'list', ('', 'd'): 'dict'}
    assert parser.definitions == {}


def test_function() -> None:
    source = (
        'def some_function():\n'
        '    """docstring"""\n'
        '    a = 1 + 1  #: comment1\n'
        '\n'
        '    b = a  #: comment2\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {}
    assert parser.definitions == {'some_function': ('def', 1, 5)}
    assert parser.deforders == {'some_function': 0}


def test_nested_function() -> None:
    source = (
        'def some_function():\n'
        '    a = 1 + 1  #: comment1\n'
        '\n'
        '    def inner_function():\n'
        '        b = 1 + 1  #: comment2\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {}
    assert parser.definitions == {'some_function': ('def', 1, 5)}
    assert parser.deforders == {'some_function': 0}


def test_class() -> None:
    source = (
        'class Foo(object):\n'
        '    attr1 = None  #: comment1\n'
        '    attr2 = None  #: comment2\n'
        '\n'
        '    def __init__(self):\n'
        '        self.a = 1 + 1  #: comment3\n'
        '        self.attr2 = 1 + 1  #: overridden\n'
        '        b = 1 + 1  #: comment5\n'
        '\n'
        '    def some_method(self):\n'
        '        c = 1 + 1  #: comment6\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('Foo', 'attr1'): 'comment1',
        ('Foo', 'a'): 'comment3',
        ('Foo', 'attr2'): 'overridden',
    }
    assert parser.definitions == {
        'Foo': ('class', 1, 11),
        'Foo.__init__': ('def', 5, 8),
        'Foo.some_method': ('def', 10, 11),
    }
    assert parser.deforders == {
        'Foo': 0,
        'Foo.attr1': 1,
        'Foo.__init__': 3,
        'Foo.a': 4,
        'Foo.attr2': 5,
        'Foo.some_method': 6,
    }


def test_class_uses_non_self() -> None:
    source = (
        'class Foo(object):\n'
        '    def __init__(this):\n'
        '        this.a = 1 + 1  #: comment\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('Foo', 'a'): 'comment'}
    assert parser.definitions == {'Foo': ('class', 1, 3), 'Foo.__init__': ('def', 2, 3)}
    assert parser.deforders == {'Foo': 0, 'Foo.__init__': 1, 'Foo.a': 2}


def test_nested_class() -> None:
    source = (
        'class Foo(object):\n'
        '    attr1 = None  #: comment1\n'
        '\n'
        '    class Bar(object):\n'
        '        attr2 = None  #: comment2\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('Foo', 'attr1'): 'comment1',
        ('Foo.Bar', 'attr2'): 'comment2',
    }
    assert parser.definitions == {'Foo': ('class', 1, 5), 'Foo.Bar': ('class', 4, 5)}
    assert parser.deforders == {
        'Foo': 0,
        'Foo.attr1': 1,
        'Foo.Bar': 2,
        'Foo.Bar.attr2': 3,
    }


def test_class_comment() -> None:
    source = (
        'import logging\n'
        'logger = logging.getLogger(__name__)\n'
        '\n'
        'class Foo(object):\n'
        '    """Bar"""\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {}
    assert parser.definitions == {'Foo': ('class', 4, 5)}


def test_comment_picker_multiline_string() -> None:
    source = (
        'class Foo(object):\n'
        '    a = None\n'
        '    """multiline\n'
        '    docstring\n'
        '    """\n'
        '    b = None\n'
        '    """\n'
        '    docstring\n'
        '    starts with::\n'
        '\n'
        '        empty line"""\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {
        ('Foo', 'a'): 'multiline\ndocstring',
        ('Foo', 'b'): 'docstring\nstarts with::\n\n    empty line',
    }


def test_decorators() -> None:
    source = (
        '@deco\n'
        'def func1(): pass\n'
        '\n'
        '@deco(param1, param2)\n'
        'def func2(): pass\n'
        '\n'
        '@deco1\n'
        '@deco2\n'
        'def func3(): pass\n'
        '\n'
        '@deco\n'
        'class Foo():\n'
        '    @deco1\n'
        '    @deco2\n'
        '    def method(self): pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.definitions == {
        'func1': ('def', 1, 2),
        'func2': ('def', 4, 5),
        'func3': ('def', 7, 9),
        'Foo': ('class', 11, 15),
        'Foo.method': ('def', 13, 15),
    }


def test_async_function_and_method() -> None:
    source = (
        'async def some_function():\n'
        '    """docstring"""\n'
        '    a = 1 + 1  #: comment1\n'
        '\n'
        'class Foo:\n'
        '    async def method(self):\n'
        '        pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.definitions == {
        'some_function': ('def', 1, 3),
        'Foo': ('class', 5, 7),
        'Foo.method': ('def', 6, 7),
    }


def test_imports() -> None:
    source = (
        'import sys\n'
        'from os import environment, path\n'
        '\n'
        'import sphinx as Sphinx\n'
        'from sphinx.application import Sphinx as App\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.definitions == {}
    assert parser.deforders == {
        'sys': 0,
        'environment': 1,
        'path': 2,
        'Sphinx': 3,
        'App': 4,
    }


def test_formfeed_char() -> None:
    source = 'class Foo:\n\f\n    attr = 1234  #: comment\n'
    parser = Parser(source)
    parser.parse()
    assert parser.comments == {('Foo', 'attr'): 'comment'}


def test_typing_final() -> None:
    source = (
        'import typing\n'
        '\n'
        '@typing.final\n'
        'def func(): pass\n'
        '\n'
        '@typing.final\n'
        'class Foo:\n'
        '    @typing.final\n'
        '    def meth(self):\n'
        '        pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.finals == ['func', 'Foo', 'Foo.meth']


def test_typing_final_from_import() -> None:
    source = (
        'from typing import final\n'
        '\n'
        '@final\n'
        'def func(): pass\n'
        '\n'
        '@final\n'
        'class Foo:\n'
        '    @final\n'
        '    def meth(self):\n'
        '        pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.finals == ['func', 'Foo', 'Foo.meth']


def test_typing_final_import_as() -> None:
    source = (
        'import typing as foo\n'
        '\n'
        '@foo.final\n'
        'def func(): pass\n'
        '\n'
        '@foo.final\n'
        'class Foo:\n'
        '    @typing.final\n'
        '    def meth(self):\n'
        '        pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.finals == ['func', 'Foo']


def test_typing_final_from_import_as() -> None:
    source = (
        'from typing import final as bar\n'
        '\n'
        '@bar\n'
        'def func(): pass\n'
        '\n'
        '@bar\n'
        'class Foo:\n'
        '    @final\n'
        '    def meth(self):\n'
        '        pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.finals == ['func', 'Foo']


def test_typing_final_not_imported() -> None:
    source = (
        '@typing.final\n'
        'def func(): pass\n'
        '\n'
        '@typing.final\n'
        'class Foo:\n'
        '    @final\n'
        '    def meth(self):\n'
        '        pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.finals == []


def test_typing_overload() -> None:
    source = (
        'import typing\n'
        '\n'
        '@typing.overload\n'
        'def func(x: int, y: int) -> int: pass\n'
        '\n'
        '@typing.overload\n'
        'def func(x: str, y: str) -> str: pass\n'
        '\n'
        'def func(x, y): pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.overloads == {
        'func': [
            signature_from_str('(x: int, y: int) -> int'),
            signature_from_str('(x: str, y: str) -> str'),
        ]
    }


def test_typing_overload_from_import() -> None:
    source = (
        'from typing import overload\n'
        '\n'
        '@overload\n'
        'def func(x: int, y: int) -> int: pass\n'
        '\n'
        '@overload\n'
        'def func(x: str, y: str) -> str: pass\n'
        '\n'
        'def func(x, y): pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.overloads == {
        'func': [
            signature_from_str('(x: int, y: int) -> int'),
            signature_from_str('(x: str, y: str) -> str'),
        ]
    }


def test_typing_overload_import_as() -> None:
    source = (
        'import typing as foo\n'
        '\n'
        '@foo.overload\n'
        'def func(x: int, y: int) -> int: pass\n'
        '\n'
        '@foo.overload\n'
        'def func(x: str, y: str) -> str: pass\n'
        '\n'
        'def func(x, y): pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.overloads == {
        'func': [
            signature_from_str('(x: int, y: int) -> int'),
            signature_from_str('(x: str, y: str) -> str'),
        ]
    }


def test_typing_overload_from_import_as() -> None:
    source = (
        'from typing import overload as bar\n'
        '\n'
        '@bar\n'
        'def func(x: int, y: int) -> int: pass\n'
        '\n'
        '@bar\n'
        'def func(x: str, y: str) -> str: pass\n'
        '\n'
        'def func(x, y): pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.overloads == {
        'func': [
            signature_from_str('(x: int, y: int) -> int'),
            signature_from_str('(x: str, y: str) -> str'),
        ]
    }


def test_typing_overload_not_imported() -> None:
    source = (
        '@typing.final\n'
        'def func(x: int, y: int) -> int: pass\n'
        '\n'
        '@typing.final\n'
        'def func(x: str, y: str) -> str: pass\n'
        '\n'
        'def func(x, y): pass\n'
    )
    parser = Parser(source)
    parser.parse()
    assert parser.overloads == {}



from __future__ import annotations

import os
import re
import sys
import traceback
from typing import TYPE_CHECKING

import pytest

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.errors import SphinxError

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp

ENV_WARNINGS = """\
{root}/autodoc_fodder.py:docstring of autodoc_fodder.MarkupError:\\d+: \
WARNING: Explicit markup ends without a blank line; unexpected unindent. \\[docutils\\]
{root}/index.rst:\\d+: WARNING: Encoding 'utf-8-sig' used for reading included \
file '{root}/wrongenc.inc' seems to be wrong, try giving an :encoding: option \\[docutils\\]
{root}/index.rst:\\d+: WARNING: invalid single index entry '' \\[index\\]
{root}/index.rst:\\d+: WARNING: image file not readable: foo.png \\[image.not_readable\\]
{root}/index.rst:\\d+: WARNING: download file not readable: {root}/nonexisting.png \\[download.not_readable\\]
{root}/undecodable.rst:\\d+: WARNING: undecodable source characters, replacing \
with "\\?": b?'here: >>>(\\\\|/)xbb<<<((\\\\|/)r)?'
"""

HTML_WARNINGS = (
    ENV_WARNINGS
    + """\
{root}/index.rst:\\d+: WARNING: unknown option: '&option' \\[ref.option\\]
{root}/index.rst:\\d+: WARNING: citation not found: missing \\[ref.ref\\]
{root}/index.rst:\\d+: WARNING: a suitable image for html builder not found: foo.\\*
{root}/index.rst:\\d+: WARNING: Lexing literal_block ".*" as "c" resulted in an error at token: ".*". Retrying in relaxed mode. \\[misc.highlighting_failure\\]
"""
)

LATEX_WARNINGS = (
    ENV_WARNINGS
    + """\
{root}/index.rst:\\d+: WARNING: unknown option: '&option' \\[ref.option\\]
{root}/index.rst:\\d+: WARNING: citation not found: missing \\[ref.ref\\]
{root}/index.rst:\\d+: WARNING: a suitable image for latex builder not found: foo.\\*
{root}/index.rst:\\d+: WARNING: Lexing literal_block ".*" as "c" resulted in an error at token: ".*". Retrying in relaxed mode. \\[misc.highlighting_failure\\]
"""
)

TEXINFO_WARNINGS = (
    ENV_WARNINGS
    + """\
{root}/index.rst:\\d+: WARNING: unknown option: '&option' \\[ref.option\\]
{root}/index.rst:\\d+: WARNING: citation not found: missing \\[ref.ref\\]
{root}/index.rst:\\d+: WARNING: a suitable image for texinfo builder not found: foo.\\*
{root}/index.rst:\\d+: WARNING: a suitable image for texinfo builder not found: \
\\['application/pdf', 'image/svg\\+xml'\\] \\(svgimg.\\*\\)
"""
)


def _check_warnings(expected_warnings: str, warning: str) -> None:
    warnings = strip_escape_sequences(re.sub(re.escape(os.sep) + '{1,2}', '/', warning))
    assert re.match(f'{expected_warnings}$', warnings), (
        "Warnings don't match:\n"
        f'--- Expected (regex):\n{expected_warnings}\n'
        f'--- Got:\n{warnings}'
    )
    sys.modules.pop('autodoc_fodder', None)


@pytest.mark.sphinx(
    'html',
    testroot='warnings',
    freshenv=True,
)
def test_html_warnings(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    warnings_exp = HTML_WARNINGS.format(root=re.escape(app.srcdir.as_posix()))
    _check_warnings(warnings_exp, app.warning.getvalue())


@pytest.mark.sphinx(
    'html',
    testroot='warnings',
    freshenv=True,
    exception_on_warning=True,
)
def test_html_warnings_exception_on_warning(app: SphinxTestApp) -> None:
    try:
        app.build(force_all=True)
        pytest.fail('Expected an exception to be raised')
    except SphinxError:
        tb = traceback.format_exc()
        assert 'unindent_warning' in tb
        assert 'pending_warnings' not in tb


@pytest.mark.sphinx(
    'latex',
    testroot='warnings',
    freshenv=True,
)
def test_latex_warnings(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    warnings_exp = LATEX_WARNINGS.format(root=re.escape(app.srcdir.as_posix()))
    _check_warnings(warnings_exp, app.warning.getvalue())


@pytest.mark.sphinx(
    'texinfo',
    testroot='warnings',
    freshenv=True,
)
def test_texinfo_warnings(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    warnings_exp = TEXINFO_WARNINGS.format(root=re.escape(app.srcdir.as_posix()))
    _check_warnings(warnings_exp, app.warning.getvalue())


def test_uncacheable_config_warning(make_app, tmp_path):
    """Test that an unpickleable config value raises a warning."""
    tmp_path.joinpath('conf.py').write_text(
        """\
my_config = lambda: None
show_warning_types = True
def setup(app):
    app.add_config_value('my_config', None, 'env')
    """,
        encoding='utf-8',
    )
    tmp_path.joinpath('index.rst').write_text('Test\n====\n', encoding='utf-8')
    app = make_app(srcdir=tmp_path)
    app.build()
    assert strip_escape_sequences(app.warning.getvalue()).strip() == (
        "WARNING: cannot cache unpickleable configuration value: 'my_config' "
        '(because it contains a function, class, or module object) [config.cache]'
    )



"""Test the only directive with the test root."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest
from docutils import nodes

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('text', testroot='directive-only')
def test_sectioning(app: SphinxTestApp) -> None:
    def getsects(section):
        if not isinstance(section, nodes.section):
            return [getsects(n) for n in section.children]
        title = section.next_node(nodes.title).astext().strip()
        subsects = []
        children = section.children[:]
        while children:
            node = children.pop(0)
            if isinstance(node, nodes.section):
                subsects.append(node)
                continue
            children = list(node.children) + children
        return [title, [getsects(subsect) for subsect in subsects]]

    def testsects(prefix, sects, indent=0):
        title = sects[0]
        parent_num = title.split()[0]
        assert prefix == parent_num, f'Section out of place: {title!r}'
        for i, subsect in enumerate(sects[1]):
            num = subsect[0].split()[0]
            assert re.match('[0-9]+[.0-9]*[.]', num), (
                f'Unnumbered section: {subsect[0]!r}'
            )
            testsects(prefix + str(i + 1) + '.', subsect, indent + 4)

    app.build(filenames=[app.srcdir / 'only.rst'])
    doctree = app.env.get_doctree('only')
    app.env.apply_post_transforms(doctree, 'only')

    parts = [getsects(n) for n in doctree.children if isinstance(n, nodes.section)]
    for i, s in enumerate(parts):
        testsects(str(i + 1) + '.', s, 4)
    actual_headings = '\n'.join(p[0] for p in parts)
    assert len(parts) == 4, (
        f'Expected 4 document level headings, got:\n{actual_headings}'
    )



"""Test the autosummary extension."""

from __future__ import annotations

import sys
from contextlib import chdir
from io import StringIO
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.ext.autosummary import (
    autosummary_table,
    autosummary_toc,
    extract_summary,
    import_by_name,
    mangle_signature,
)
from sphinx.ext.autosummary.generate import (
    AutosummaryEntry,
    generate_autosummary_content,
    generate_autosummary_docs,
)
from sphinx.ext.autosummary.generate import main as autogen_main
from sphinx.testing.util import assert_node, etree_parse
from sphinx.util.docutils import new_document

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element

html_warnfile = StringIO()


defaults = {
    'extensions': ['sphinx.ext.autosummary'],
    'autosummary_generate': True,
    'autosummary_generate_overwrite': False,
}


@pytest.fixture(autouse=True)
def _unload_target_module():
    sys.modules.pop('target', None)


def test_mangle_signature() -> None:
    TEST_SIGNATURE = """
    () :: ()
    (a, b, c, d, e) :: (a, b, c, d, e)
    (a, b, c=1, d=2, e=3) :: (a, b[, c, d, e])
    (a, b, aaa=1, bbb=1, ccc=1, eee=1, fff=1, ggg=1, hhh=1, iii=1, jjj=1)\
    :: (a, b[, aaa, bbb, ccc, ...])
    (a, b, c=(), d=<foo>) :: (a, b[, c, d])
    (a, b, c='foobar()', d=123) :: (a, b[, c, d])
    (a, b[, c]) :: (a, b[, c])
    (a, b[, cxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx]) :: (a, b[, ...)
    (a, b='c=d, e=f, g=h', c=3) :: (a[, b, c])
    (a, b="c=d, e=f, g=h", c=3) :: (a[, b, c])
    (a, b='c=d, \\'e=f,\\' g=h', c=3) :: (a[, b, c])
    (a, b='c=d, ', e='\\\\' g=h, c=3) :: (a[, b, e, c])
    (a, b={'c=d, ': 3, '\\\\': 3}) :: (a[, b])
    (a=1, b=2, c=3) :: ([a, b, c])
    (a=1, b=<SomeClass: a, b, c>, c=3) :: ([a, b, c])
    (a=1, b=T(a=1, b=2), c=3) :: ([a, b, c])
    (a: Tuple[int, str], b: int) -> str :: (a, b)
    """

    TEST = [
        list(map(str.strip, x.split('::')))
        for x in TEST_SIGNATURE.split('\n')
        if '::' in x
    ]
    for inp, outp in TEST:
        res = mangle_signature(inp).strip().replace('\u00a0', ' ')
        assert res == outp, f"'{inp}' -> '{res}' != '{outp}'"


def test_extract_summary(capsys):
    settings = Mock(
        language_code='en',
        id_prefix='',
        auto_id_prefix='',
        pep_reference=False,
        rfc_reference=False,
    )
    document = new_document('', settings)

    # normal case
    doc = [
        '',
        'This is a first sentence. And second one.',
        '',
        'Second block is here',
    ]
    assert extract_summary(doc, document) == 'This is a first sentence.'

    # inliner case
    doc = [
        'This sentence contains *emphasis text having dots.*,',
        'it does not break sentence.',
    ]
    assert extract_summary(doc, document) == ' '.join(doc)

    # abbreviations
    doc = ['Blabla, i.e. bla.']
    assert extract_summary(doc, document) == ' '.join(doc)

    doc = ['Blabla, (i.e. bla).']
    assert extract_summary(doc, document) == ' '.join(doc)

    doc = ['Blabla, e.g. bla.']
    assert extract_summary(doc, document) == ' '.join(doc)

    doc = ['Blabla, (e.g. bla).']
    assert extract_summary(doc, document) == ' '.join(doc)

    doc = ['Blabla, et al. bla.']
    assert extract_summary(doc, document) == ' '.join(doc)

    # literal
    doc = ['blah blah::']
    assert extract_summary(doc, document) == 'blah blah.'

    # heading
    doc = [
        'blah blah',
        '=========',
    ]
    assert extract_summary(doc, document) == 'blah blah'

    doc = [
        '=========',
        'blah blah',
        '=========',
    ]
    assert extract_summary(doc, document) == 'blah blah'

    # hyperlink target
    doc = ['Do `this <https://www.sphinx-doc.org/>`_ and that. blah blah blah.']
    extracted = extract_summary(doc, document)
    assert extracted == 'Do `this <https://www.sphinx-doc.org/>`_ and that.'

    _, err = capsys.readouterr()
    assert err == ''


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-ext',
    confoverrides=defaults.copy(),
    copy_test_root=True,
)
def test_get_items_summary(make_app, app_params):
    import sphinx.ext.autosummary
    import sphinx.ext.autosummary.generate

    args, kwargs = app_params
    app = make_app(*args, **kwargs)
    sphinx.ext.autosummary.generate.setup_documenters(app)
    # monkey-patch Autosummary.get_items so we can easily get access to it's
    # results..
    orig_get_items = sphinx.ext.autosummary.Autosummary.get_items

    autosummary_items = {}

    def new_get_items(self, names, *args, **kwargs):
        results = orig_get_items(self, names, *args, **kwargs)
        for name, result in zip(names, results, strict=True):
            autosummary_items[name] = result  # NoQA: PERF403
        return results

    def handler(app, what, name, obj, options, lines):
        assert isinstance(lines, list)

        # ensure no docstring is processed twice:
        assert 'THIS HAS BEEN HANDLED' not in lines
        lines.append('THIS HAS BEEN HANDLED')

    app.connect('autodoc-process-docstring', handler)

    sphinx.ext.autosummary.Autosummary.get_items = new_get_items
    try:
        app.build(force_all=True)
    finally:
        sphinx.ext.autosummary.Autosummary.get_items = orig_get_items

    html_warnings = app._warning.getvalue()
    assert html_warnings == ''

    expected_values = {
        'with_sentence': 'I have a sentence which spans multiple lines.',
        'no_sentence': "this doesn't start with a capital.",
        'empty_line': 'This is the real summary',
        'module_attr': 'This is a module attribute',
        'C.class_attr': 'This is a class attribute',
        'C.instance_attr': 'This is an instance attribute',
        'C.prop_attr1': 'This is a function docstring',
        'C.prop_attr2': 'This is a attribute docstring',
        'C.C2': 'This is a nested inner class docstring',
    }
    for key, expected in expected_values.items():
        assert autosummary_items[key][2] == expected, (
            f'Summary for {key} was {autosummary_items[key]!r} - expected {expected!r}'
        )

    # check an item in detail
    assert 'func' in autosummary_items
    func_attrs = (
        'func',
        '(arg_, *args, **kwargs)',
        'Test function take an argument ended with underscore.',
        'dummy_module.func',
    )
    assert autosummary_items['func'] == func_attrs


def str_content(elem: Element) -> str:
    if elem.text is not None:
        return elem.text
    else:
        return ''.join(str_content(e) for e in elem)


@pytest.mark.sphinx(
    'xml',
    testroot='ext-autosummary-ext',
    confoverrides=defaults.copy(),
    copy_test_root=True,
)
def test_escaping(app):
    app.build(force_all=True)

    docpage = app.builder.outdir / 'underscore_module_.xml'
    assert docpage.exists()

    title = etree_parse(docpage).find('section/title')
    assert str_content(title) == 'underscore_module_'


@pytest.mark.sphinx('html', testroot='ext-autosummary', copy_test_root=True)
def test_autosummary_generate_content_for_module(app):
    import autosummary_dummy_module  # type: ignore[import-not-found]

    template = Mock()

    generate_autosummary_content(
        'autosummary_dummy_module',
        autosummary_dummy_module,
        None,
        template,
        None,
        False,
        False,
        {},
        config=app.config,
        events=app.events,
        registry=app.registry,
    )
    assert template.render.call_args[0][0] == 'module'

    context = template.render.call_args[0][1]
    assert context['members'] == [
        'CONSTANT1',
        'CONSTANT2',
        'Exc',
        'Foo',
        '_Baz',
        '_Exc',
        '__all__',
        '__builtins__',
        '__cached__',
        '__doc__',
        '__file__',
        '__name__',
        '__package__',
        '_quux',
        'bar',
        'non_imported_member',
        'quuz',
        'qux',
    ]
    assert context['functions'] == ['bar']
    assert context['all_functions'] == ['_quux', 'bar']
    assert context['classes'] == ['Foo']
    assert context['all_classes'] == ['Foo', '_Baz']
    assert context['exceptions'] == ['Exc']
    assert context['all_exceptions'] == ['Exc', '_Exc']
    assert context['attributes'] == ['CONSTANT1', 'qux', 'quuz', 'non_imported_member']
    assert context['all_attributes'] == [
        'CONSTANT1',
        'qux',
        'quuz',
        'non_imported_member',
    ]
    assert context['fullname'] == 'autosummary_dummy_module'
    assert context['module'] == 'autosummary_dummy_module'
    assert context['objname'] == ''
    assert context['name'] == ''
    assert context['objtype'] == 'module'


@pytest.mark.sphinx('html', testroot='ext-autosummary', copy_test_root=True)
def test_autosummary_generate_content_for_module___all__(app):
    import autosummary_dummy_module

    template = Mock()
    app.config.autosummary_ignore_module_all = False

    generate_autosummary_content(
        'autosummary_dummy_module',
        autosummary_dummy_module,
        None,
        template,
        None,
        False,
        False,
        {},
        config=app.config,
        events=app.events,
        registry=app.registry,
    )
    assert template.render.call_args[0][0] == 'module'

    context = template.render.call_args[0][1]
    assert context['members'] == [
        'CONSTANT1',
        'Exc',
        'Foo',
        '_Baz',
        'bar',
        'qux',
        'path',
    ]
    assert context['functions'] == ['bar']
    assert context['all_functions'] == ['bar']
    assert context['classes'] == ['Foo']
    assert context['all_classes'] == ['Foo', '_Baz']
    assert context['exceptions'] == ['Exc']
    assert context['all_exceptions'] == ['Exc']
    assert context['attributes'] == ['CONSTANT1', 'qux']
    assert context['all_attributes'] == ['CONSTANT1', 'qux']
    assert context['fullname'] == 'autosummary_dummy_module'
    assert context['module'] == 'autosummary_dummy_module'
    assert context['objname'] == ''
    assert context['name'] == ''
    assert context['objtype'] == 'module'


@pytest.mark.sphinx('html', testroot='ext-autosummary', copy_test_root=True)
def test_autosummary_generate_content_for_module_skipped(app):
    import autosummary_dummy_module

    template = Mock()

    def skip_member(app, what, name, obj, skip, options):
        if name in {'Foo', 'bar', 'Exc'}:
            return True
        return None

    app.connect('autodoc-skip-member', skip_member)
    generate_autosummary_content(
        'autosummary_dummy_module',
        autosummary_dummy_module,
        None,
        template,
        None,
        False,
        False,
        {},
        config=app.config,
        events=app.events,
        registry=app.registry,
    )
    context = template.render.call_args[0][1]
    assert context['members'] == [
        'CONSTANT1',
        'CONSTANT2',
        '_Baz',
        '_Exc',
        '__all__',
        '__builtins__',
        '__cached__',
        '__doc__',
        '__file__',
        '__name__',
        '__package__',
        '_quux',
        'non_imported_member',
        'quuz',
        'qux',
    ]
    assert context['functions'] == []
    assert context['classes'] == []
    assert context['exceptions'] == []


@pytest.mark.sphinx('html', testroot='ext-autosummary', copy_test_root=True)
def test_autosummary_generate_content_for_module_imported_members(app):
    import autosummary_dummy_module

    template = Mock()

    generate_autosummary_content(
        'autosummary_dummy_module',
        autosummary_dummy_module,
        None,
        template,
        None,
        True,
        False,
        {},
        config=app.config,
        events=app.events,
        registry=app.registry,
    )
    assert template.render.call_args[0][0] == 'module'

    context = template.render.call_args[0][1]
    assert context['members'] == [
        'CONSTANT1',
        'CONSTANT2',
        'Class',
        'Exc',
        'Foo',
        'Union',
        '_Baz',
        '_Exc',
        '__all__',
        '__builtins__',
        '__cached__',
        '__doc__',
        '__file__',
        '__loader__',
        '__name__',
        '__package__',
        '__spec__',
        '_quux',
        'bar',
        'considered_as_imported',
        'non_imported_member',
        'path',
        'quuz',
        'qux',
    ]
    assert context['functions'] == ['bar']
    assert context['all_functions'] == ['_quux', 'bar']
    assert context['classes'] == ['Class', 'Foo']
    assert context['all_classes'] == ['Class', 'Foo', '_Baz']
    assert context['exceptions'] == ['Exc']
    assert context['all_exceptions'] == ['Exc', '_Exc']
    assert context['attributes'] == ['CONSTANT1', 'qux', 'quuz', 'non_imported_member']
    assert context['all_attributes'] == [
        'CONSTANT1',
        'qux',
        'quuz',
        'non_imported_member',
    ]
    assert context['fullname'] == 'autosummary_dummy_module'
    assert context['module'] == 'autosummary_dummy_module'
    assert context['objname'] == ''
    assert context['name'] == ''
    assert context['objtype'] == 'module'


@pytest.mark.sphinx('html', testroot='ext-autosummary', copy_test_root=True)
def test_autosummary_generate_content_for_module_imported_members_inherited_module(app):
    import autosummary_dummy_inherited_module  # type: ignore[import-not-found]

    template = Mock()

    generate_autosummary_content(
        'autosummary_dummy_inherited_module',
        autosummary_dummy_inherited_module,
        None,
        template,
        None,
        True,
        False,
        {},
        config=app.config,
        events=app.events,
        registry=app.registry,
    )
    assert template.render.call_args[0][0] == 'module'

    context = template.render.call_args[0][1]
    assert context['members'] == [
        'Foo',
        'InheritedAttrClass',
        '__all__',
        '__builtins__',
        '__cached__',
        '__doc__',
        '__file__',
        '__loader__',
        '__name__',
        '__package__',
        '__spec__',
    ]
    assert context['functions'] == []
    assert context['classes'] == ['Foo', 'InheritedAttrClass']
    assert context['exceptions'] == []
    assert context['all_exceptions'] == []
    assert context['attributes'] == []
    assert context['all_attributes'] == []
    assert context['fullname'] == 'autosummary_dummy_inherited_module'
    assert context['module'] == 'autosummary_dummy_inherited_module'
    assert context['objname'] == ''
    assert context['name'] == ''
    assert context['objtype'] == 'module'


@pytest.mark.sphinx('dummy', testroot='ext-autosummary', copy_test_root=True)
def test_autosummary_generate(app):
    app.build(force_all=True)

    doctree = app.env.get_doctree('index')
    assert_node(
        doctree,
        (
            nodes.paragraph,
            nodes.paragraph,
            addnodes.tabular_col_spec,
            autosummary_table,
            [autosummary_toc, addnodes.toctree],
        ),
    )
    assert_node(
        doctree[3],
        [
            autosummary_table,
            nodes.table,
            nodes.tgroup,
            (
                nodes.colspec,
                nodes.colspec,
                [
                    nodes.tbody,
                    (
                        nodes.row,
                        nodes.row,
                        nodes.row,
                        nodes.row,
                        nodes.row,
                        nodes.row,
                        nodes.row,
                        nodes.row,
                    ),
                ],
            ),
        ],
    )
    assert_node(doctree[4][0], addnodes.toctree, caption='An autosummary')

    assert len(doctree[3][0][0][2]) == 8
    assert doctree[3][0][0][2][0].astext() == 'autosummary_dummy_module\n\n'
    assert doctree[3][0][0][2][1].astext() == 'autosummary_dummy_module.Foo()\n\n'
    assert doctree[3][0][0][2][2].astext() == 'autosummary_dummy_module.Foo.Bar()\n\n'
    assert doctree[3][0][0][2][3].astext() == (
        'autosummary_dummy_module.Foo.value\n\ndocstring'
    )
    assert doctree[3][0][0][2][4].astext() == 'autosummary_dummy_module.bar(x[, y])\n\n'
    assert doctree[3][0][0][2][5].astext() == (
        'autosummary_dummy_module.qux\n\na module-level attribute'
    )
    assert doctree[3][0][0][2][6].astext() == (
        'autosummary_dummy_inherited_module.InheritedAttrClass()\n\n'
    )
    assert doctree[3][0][0][2][7].astext() == (
        'autosummary_dummy_inherited_module.InheritedAttrClass.subclassattr\n\nother docstring'
    )

    path = app.srcdir / 'generated' / 'autosummary_dummy_module.rst'
    module = path.read_text(encoding='utf8')

    assert '   .. autosummary::\n   \n      Foo\n   \n' in module
    assert (
        '   .. autosummary::\n'
        '   \n'
        '      CONSTANT1\n'
        '      qux\n'
        '      quuz\n'
        '      non_imported_member\n'
        '   \n'
    ) in module

    path = app.srcdir / 'generated' / 'autosummary_dummy_module.Foo.rst'
    Foo = path.read_text(encoding='utf8')
    assert '.. automethod:: __init__' in Foo
    assert (
        '   .. autosummary::\n   \n      ~Foo.__init__\n      ~Foo.bar\n   \n'
    ) in Foo
    assert (
        '   .. autosummary::\n'
        '   \n'
        '      ~Foo.CONSTANT3\n'
        '      ~Foo.CONSTANT4\n'
        '      ~Foo.baz\n'
        '      ~Foo.value\n'
        '   \n'
    ) in Foo

    path = app.srcdir / 'generated' / 'autosummary_dummy_module.Foo.Bar.rst'
    FooBar = path.read_text(encoding='utf8')
    assert (
        '.. currentmodule:: autosummary_dummy_module\n\n.. autoclass:: Foo.Bar\n'
    ) in FooBar

    path = app.srcdir / 'generated' / 'autosummary_dummy_module.Foo.value.rst'
    Foo_value = path.read_text(encoding='utf8')
    assert (
        '.. currentmodule:: autosummary_dummy_module\n\n.. autoattribute:: Foo.value'
    ) in Foo_value

    path = app.srcdir / 'generated' / 'autosummary_dummy_module.qux.rst'
    qux = path.read_text(encoding='utf8')
    assert '.. currentmodule:: autosummary_dummy_module\n\n.. autodata:: qux' in qux

    path = (
        app.srcdir
        / 'generated'
        / 'autosummary_dummy_inherited_module.InheritedAttrClass.rst'
    )
    InheritedAttrClass = path.read_text(encoding='utf8')
    print(InheritedAttrClass)
    assert '.. automethod:: __init__' in Foo
    assert (
        '   .. autosummary::\n'
        '   \n'
        '      ~InheritedAttrClass.__init__\n'
        '      ~InheritedAttrClass.bar\n'
        '   \n'
    ) in InheritedAttrClass
    assert (
        '   .. autosummary::\n'
        '   \n'
        '      ~InheritedAttrClass.CONSTANT3\n'
        '      ~InheritedAttrClass.CONSTANT4\n'
        '      ~InheritedAttrClass.baz\n'
        '      ~InheritedAttrClass.subclassattr\n'
        '      ~InheritedAttrClass.value\n'
        '   \n'
    ) in InheritedAttrClass

    path = (
        app.srcdir
        / 'generated'
        / 'autosummary_dummy_inherited_module.InheritedAttrClass.subclassattr.rst'
    )
    InheritedAttrClass_subclassattr = path.read_text(encoding='utf8')
    assert (
        '.. currentmodule:: autosummary_dummy_inherited_module\n'
        '\n'
        '.. autoattribute:: InheritedAttrClass.subclassattr'
    ) in InheritedAttrClass_subclassattr


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary',
    confoverrides={'autosummary_generate_overwrite': False},
    copy_test_root=True,
)
def test_autosummary_generate_overwrite1(app_params, make_app):
    args, kwargs = app_params
    srcdir = kwargs.get('srcdir')

    (srcdir / 'generated').mkdir(parents=True, exist_ok=True)
    (srcdir / 'generated' / 'autosummary_dummy_module.rst').write_text('', encoding='utf8')  # fmt: skip

    app = make_app(*args, **kwargs)
    path = srcdir / 'generated' / 'autosummary_dummy_module.rst'
    content = path.read_text(encoding='utf8')
    assert content == ''
    assert 'autosummary_dummy_module.rst' not in app._warning.getvalue()


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary',
    confoverrides={'autosummary_generate_overwrite': True},
    copy_test_root=True,
)
def test_autosummary_generate_overwrite2(app_params, make_app):
    args, kwargs = app_params
    srcdir = kwargs.get('srcdir')

    (srcdir / 'generated').mkdir(parents=True, exist_ok=True)
    (srcdir / 'generated' / 'autosummary_dummy_module.rst').write_text('', encoding='utf8')  # fmt: skip

    app = make_app(*args, **kwargs)
    path = srcdir / 'generated' / 'autosummary_dummy_module.rst'
    content = path.read_text(encoding='utf8')
    assert content != ''
    assert 'autosummary_dummy_module.rst' not in app._warning.getvalue()


@pytest.mark.sphinx('dummy', testroot='ext-autosummary-recursive', copy_test_root=True)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_autosummary_recursive(app):
    sys.modules.pop('package', None)  # unload target module to clear the module cache

    app.build()

    # autosummary having :recursive: option
    assert (app.srcdir / 'generated' / 'package.rst').exists()
    assert (app.srcdir / 'generated' / 'package.module.rst').exists()
    assert not (app.srcdir / 'generated' / 'package.module_importfail.rst').exists()
    assert (app.srcdir / 'generated' / 'package.package.rst').exists()
    assert (app.srcdir / 'generated' / 'package.package.module.rst').exists()

    # autosummary not having :recursive: option
    assert (app.srcdir / 'generated' / 'package2.rst').exists()
    assert not (app.srcdir / 'generated' / 'package2.module.rst').exists()

    # Check content of recursively generated stub-files
    content = (app.srcdir / 'generated' / 'package.rst').read_text(encoding='utf8')
    assert 'module' in content
    assert 'package' in content
    assert 'module_importfail' in content
    # we no longer generate fully-qualified module names.
    assert 'package.module' not in content
    assert 'package.package' not in content
    assert 'package.module_importfail' not in content

    path = app.srcdir / 'generated' / 'package.package.rst'
    content = path.read_text(encoding='utf8')
    assert 'module' in content
    assert 'package.package.module' not in content

    warnings = app.warning.getvalue()
    assert 'Summarised items should not include the current module.' not in warnings


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-recursive',
    srcdir='test_autosummary_recursive_skips_mocked_modules',
    confoverrides={'autosummary_mock_imports': ['package.package']},
)
@pytest.mark.usefixtures('rollback_sysmodules')
def test_autosummary_recursive_skips_mocked_modules(app):
    sys.modules.pop('package', None)  # unload target module to clear the module cache
    app.build()

    assert (app.srcdir / 'generated' / 'package.rst').exists()
    assert (app.srcdir / 'generated' / 'package.module.rst').exists()
    assert not (app.srcdir / 'generated' / 'package.package.rst').exists()
    assert not (app.srcdir / 'generated' / 'package.package.module.rst').exists()


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-filename-map',
    copy_test_root=True,
)
def test_autosummary_filename_map(app):
    app.build()

    assert (app.srcdir / 'generated' / 'module_mangled.rst').exists()
    assert not (app.srcdir / 'generated' / 'autosummary_dummy_module.rst').exists()
    assert (app.srcdir / 'generated' / 'bar.rst').exists()
    assert not (app.srcdir / 'generated' / 'autosummary_dummy_module.bar.rst').exists()
    assert (app.srcdir / 'generated' / 'autosummary_dummy_module.Foo.rst').exists()

    html_warnings = app._warning.getvalue()
    assert html_warnings == ''


@pytest.mark.sphinx(
    'latex',
    testroot='ext-autosummary-ext',
    confoverrides=defaults.copy(),
    copy_test_root=True,
)
def test_autosummary_latex_table_colspec(app):
    app.build(force_all=True)
    result = (app.outdir / 'projectnamenotset.tex').read_text(encoding='utf8')
    print(app.status.getvalue())
    print(app.warning.getvalue())
    assert r'\begin{longtable}{\X{1}{2}\X{1}{2}}' in result
    assert r'p{0.5\linewidth}' not in result


def test_import_by_name() -> None:
    import sphinx
    import sphinx.ext.autosummary

    prefixed_name, obj, parent, modname = import_by_name('sphinx')
    assert prefixed_name == 'sphinx'
    assert obj is sphinx
    assert parent is None
    assert modname == 'sphinx'

    prefixed_name, obj, parent, modname = import_by_name(
        'sphinx.ext.autosummary.__name__'
    )
    assert prefixed_name == 'sphinx.ext.autosummary.__name__'
    assert obj is sphinx.ext.autosummary.__name__
    assert parent is sphinx.ext.autosummary
    assert modname == 'sphinx.ext.autosummary'

    prefixed_name, obj, parent, modname = import_by_name(
        'sphinx.ext.autosummary.Autosummary.get_items'
    )
    assert prefixed_name == 'sphinx.ext.autosummary.Autosummary.get_items'
    assert obj == sphinx.ext.autosummary.Autosummary.get_items
    assert parent is sphinx.ext.autosummary.Autosummary
    assert modname == 'sphinx.ext.autosummary'


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-mock_imports',
    copy_test_root=True,
)
def test_autosummary_mock_imports(app):
    try:
        app.build()
        assert app.warning.getvalue() == ''

        # generated/foo is generated successfully
        assert app.env.get_doctree('generated/foo')
    finally:
        sys.modules.pop('foo', None)  # unload foo module


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-imported_members',
    copy_test_root=True,
)
def test_autosummary_imported_members(app):
    try:
        app.build()
        # generated/foo is generated successfully
        assert app.env.get_doctree('generated/autosummary_dummy_package')

        path = app.srcdir / 'generated' / 'autosummary_dummy_package.rst'
        module = path.read_text(encoding='utf8')
        assert '   .. autosummary::\n   \n      Bar\n   ' in module
        assert '   .. autosummary::\n   \n      foo\n   ' in module
    finally:
        sys.modules.pop('autosummary_dummy_package', None)


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-module_all',
    copy_test_root=True,
)
def test_autosummary_module_all(app):
    try:
        app.build()
        # generated/foo is generated successfully
        assert app.env.get_doctree('generated/autosummary_dummy_package_all')
        path = app.srcdir / 'generated' / 'autosummary_dummy_package_all.rst'
        module = path.read_text(encoding='utf8')
        assert '   .. autosummary::\n   \n      PublicBar\n   \n' in module
        assert (
            '   .. autosummary::\n   \n      public_foo\n      public_baz\n   \n'
        ) in module
        assert (
            '.. autosummary::\n   :toctree:\n   :recursive:\n\n   extra_dummy_module\n'
        ) in module
    finally:
        sys.modules.pop('autosummary_dummy_package_all', None)


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary-module_empty_all',
    copy_test_root=True,
)
def test_autosummary_module_empty_all(app):
    try:
        app.build()
        # generated/foo is generated successfully
        assert app.env.get_doctree('generated/autosummary_dummy_package_empty_all')
        path = app.srcdir / 'generated' / 'autosummary_dummy_package_empty_all.rst'
        module = path.read_text(encoding='utf8')
        assert '.. automodule:: autosummary_dummy_package_empty_all' in module
        # for __all__ = (), the output should not contain any variables
        assert '__all__' not in module
        assert '__builtins__' not in module
        assert '__cached__' not in module
        assert '__doc__' not in module
        assert '__file__' not in module
        assert '__loader__' not in module
        assert '__name__' not in module
        assert '__package__' not in module
        assert '__path__' not in module
        assert '__spec__' not in module
    finally:
        sys.modules.pop('autosummary_dummy_package_all', None)


@pytest.mark.sphinx(
    'html',
    testroot='ext-autodoc',
    confoverrides={'extensions': ['sphinx.ext.autosummary']},
    copy_test_root=True,
)
def test_generate_autosummary_docs_property(app):
    with patch('sphinx.ext.autosummary.generate.find_autosummary_in_files') as mock:
        mock.return_value = [
            AutosummaryEntry('target.methods.Base.prop', 'prop', None, False)
        ]
        generate_autosummary_docs([], output_dir=app.srcdir, app=app)

    content = (app.srcdir / 'target.methods.Base.prop.rst').read_text(encoding='utf8')
    assert content == (
        'target.methods.Base.prop\n'
        '========================\n'
        '\n'
        '.. currentmodule:: target.methods\n'
        '\n'
        '.. autoproperty:: Base.prop'
    )


@pytest.mark.sphinx(
    'html',
    testroot='ext-autosummary-skip-member',
    copy_test_root=True,
)
def test_autosummary_skip_member(app):
    app.build()

    content = (app.srcdir / 'generate' / 'target.Foo.rst').read_text(encoding='utf8')
    assert 'Foo.skipmeth' not in content
    assert 'Foo._privatemeth' in content


@pytest.mark.sphinx('html', testroot='ext-autosummary-template', copy_test_root=True)
def test_autosummary_template(app):
    app.build()

    content = (app.srcdir / 'generate' / 'target.Foo.rst').read_text(encoding='utf8')
    assert 'EMPTY' in content


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary',
    confoverrides={'autosummary_generate': []},
)
def test_empty_autosummary_generate(app):
    app.build()
    assert (
        'WARNING: autosummary: failed to import autosummary_importfail'
    ) in app.warning.getvalue()


@pytest.mark.sphinx(
    'dummy',
    testroot='ext-autosummary',
    confoverrides={'autosummary_generate': ['unknown']},
)
def test_invalid_autosummary_generate(app):
    assert (
        'WARNING: autosummary_generate: file not found: unknown.rst'
    ) in app.warning.getvalue()


def test_autogen(rootdir, tmp_path):
    with chdir(rootdir / 'test-templating'):
        args = ['-o', str(tmp_path), '-t', '.', 'autosummary_templating.txt']
        autogen_main(args)
        assert (tmp_path / 'sphinx.application.TemplateBridge.rst').exists()


def test_autogen_remove_old(rootdir, tmp_path):
    """Test the ``--remove-old`` option."""
    tmp_path.joinpath('other.rst').write_text('old content', encoding='utf-8')
    with chdir(rootdir / 'test-templating'):
        args = ['-o', str(tmp_path), '-t', '.', 'autosummary_templating.txt']
        autogen_main(args)
        assert set(tmp_path.iterdir()) == {
            tmp_path / 'sphinx.application.TemplateBridge.rst',
            tmp_path / 'other.rst',
        }
        autogen_main([*args, '--remove-old'])
        assert set(tmp_path.iterdir()) == {
            tmp_path / 'sphinx.application.TemplateBridge.rst'
        }
